#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide **string munging** utility unit tests.

This submodule unit tests the public API of the private
:mod:`beartype._util.text.utiltextmunge` submodule.
'''

# ....................{ IMPORTS                            }....................
# !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
# !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

# ....................{ TESTS ~ case                       }....................
def test_uppercase_str_char_first():
    '''
    Test the :func:`beartype._util.text.utiltextmunge.uppercase_str_char_first`
    function.
    '''

    # Defer test-specific imports.
    from beartype._util.text.utiltextmunge import uppercase_str_char_first
    from pytest import raises

    # Assert this munger returns the expected string when passed a string.
    assert uppercase_str_char_first(text='beArTyPe') == "BeArTyPe"
    assert uppercase_str_char_first(text="<bear>") == "<bear>"
    assert uppercase_str_char_first(text="") == ""

    # Assert this munger raises the expected exception when passed a non-string.
    with raises(AssertionError):
        uppercase_str_char_first(7)

# ....................{ TESTS ~ number                     }....................
def test_number_str_lines():
    '''
    Test the :func:`beartype._util.text.utiltextmunge.number_str_lines`
    function.
    '''

    # Defer test-specific imports.
    from beartype._util.text.utiltextmunge import number_str_lines
    from pytest import raises
    from re import search

    with raises(AssertionError):
        number_str_lines(7)

    # No lines to split.
    assert number_str_lines(text="") == ""

    NEWLINE_COUNT = 20
    base_string = 'bears, beats, battlestar galactica'
    total_string = f'{base_string}\n' * NEWLINE_COUNT

    numbered_lines_string = number_str_lines(total_string)
    numbered_lines_string = numbered_lines_string.splitlines()

    # Confirm the function preserves newlines as is.
    assert len(numbered_lines_string) == NEWLINE_COUNT

    # Confirm the base string is prefixed with something.
    for line_number in range(NEWLINE_COUNT):
        assert search(
            pattern=fr'(?<!^){base_string}$',
            string=numbered_lines_string[line_number],
        ) is not None

# ....................{ TESTS ~ replace                    }....................
def test_replace_str_substrs():
    '''
    Test the :func:`beartype._util.text.utiltextmunge.replace_str_substrs`
    function.
    '''

    # Defer test-specific imports.
    from beartype._util.text.utiltextmunge import replace_str_substrs
    from beartype.roar._roarexc import _BeartypeUtilTextException
    from pytest import raises

    with raises(AssertionError):
        replace_str_substrs(text=7, old='Oh No!', new='A non-str value')

    with raises(AssertionError):
        replace_str_substrs(text='Oh No!', old=7.0, new='is a non-str value')

    with raises(AssertionError):
        replace_str_substrs(text='Oh No!', old='A non-str value of ', new=7)

    with raises(_BeartypeUtilTextException):
        replace_str_substrs(
            text="OG Beartype",
            old="I DON'T EXIST!",
            new="Therefore I do not think.",
        )

    assert replace_str_substrs(
        text="I do not think, therefore I am.",
        old="do not think",
        new="think",
    ) == "I think, therefore I am."

# ....................{ TESTS ~ [pre|suf]fix               }....................
def test_suffix_str_unless_suffixed():
    '''
    Test the
    :func:`beartype._util.text.utiltextmunge.suffix_str_unless_suffixed`
    function.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype._util.text.utiltextmunge import suffix_str_unless_suffixed
    from pytest import raises

    # ....................{ LOCALS                         }....................
    not_suffixed_text = "somefile"
    suffix = ".py"

    # ....................{ PASS                           }....................
    suffixed_text = suffix_str_unless_suffixed(
        text=not_suffixed_text, suffix=suffix)

    # Confirm text is suffixed
    assert suffixed_text == "somefile.py"

    # Attempt to reapply the suffix
    suffix_str_unless_suffixed(text=suffixed_text, suffix=suffix)

    # Confirm the text is unchanged
    assert suffixed_text == "somefile.py"

    # ....................{ FAIL                           }....................
    with raises(AssertionError):
        suffix_str_unless_suffixed(text=7, suffix="That's not a string!")

    with raises(AssertionError):
        suffix_str_unless_suffixed(text="HEY! You're not a string!", suffix=7)

# ....................{ TESTS ~ truncate                   }....................
def test_truncate_str():
    '''
    Test the :func:`beartype._util.text.utiltextmunge.truncate_str` function.
    '''

    # Defer test-specific imports.
    from beartype._util.text.utiltextmunge import truncate_str

    # Assert that truncating the empty string returns the empty string.
    assert truncate_str(text='') == ''
    assert truncate_str(text='', max_len=0) == ''

    # Assert that truncating extremely short strings (i.e., less than four
    # characters in length) truncates those strings to the desired maximum
    # length *WITHOUT* injecting an ellipsis into that truncation.
    assert truncate_str(text='Se',   max_len=1) == 'S'
    assert truncate_str(text='Sen',  max_len=2) == 'Se'
    assert truncate_str(text='Sent', max_len=3) == 'Sen'

    # Assert that truncating fairly short strings (i.e., less than seven
    # characters in length but more than four characters in length) containing
    # *NO* punctuation truncates those strings to the desired maximum length by
    # replacing all trailing characters in those strings with an ellipsis.
    assert truncate_str(text='Sent ',   max_len=4) == 'S...'
    assert truncate_str(text='Sent t',  max_len=5) == 'Se...'
    assert truncate_str(text='Sent to', max_len=6) == 'Sen...'

    # Assert that truncating fairly short strings containing one non-period
    # punctuation character truncates those strings to the desired maximum
    # length by replacing all trailing characters in those strings up to but
    # *NOT* including that non-period punctuation character with an ellipsis.
    assert truncate_str(text='Sent>',   max_len=4) == '...>'
    assert truncate_str(text='Sent >',  max_len=5) == 'S...>'
    assert truncate_str(text='Sent t>', max_len=6) == 'Se...>'

    # Assert that truncating a fairly short string containing two or more
    # non-period punctuation characters to a length larger than the length of
    # an ellipsis (i.e., 3 characters) plus the length of those non-period
    # punctuation characters (e.g., 2 in this case) truncates that string in a
    # reasonable but complex manner too complex to detail here. *sigh* 
    assert truncate_str(text='Sent/>', max_len=4) == '...>'

    # Assert that truncating fairly short strings containing two or more
    # non-period punctuation characters truncates those strings to the desired
    # maximum length by replacing all trailing characters in those strings up to
    # but *NOT* including those non-period punctuation characters with an
    # ellipsis.
    assert truncate_str(text='Sent />',   max_len=5) == '.../>'
    assert truncate_str(text='Sent t/>',  max_len=6) == 'S.../>'
    assert truncate_str(text='Sent to/>', max_len=7) == 'Se.../>'

    # Assert that truncating fairly short strings suffixed by one period
    # followed by two or more non-period punctuation characters truncates those
    # strings to the desired maximum length by replacing all trailing characters
    # in those strings up to and including those periods but *NOT* those
    # non-period punctuation characters with an ellipsis.
    assert truncate_str(text='Sen."?',    max_len=5) == '..."?'
    assert truncate_str(text='Sent."?',   max_len=6) == 'S..."?'
    assert truncate_str(text='Sent ."?',  max_len=7) == 'Se..."?'
    assert truncate_str(text='Sent t."?', max_len=8) == 'Sen..."?'

    # Assert that truncating fairly short strings suffixed by an ellipsis
    # truncates those strings to the desired maximum length while preserving
    # that ellipsis as is.
    assert truncate_str(text='Se...',   max_len=4) == 'S...'
    assert truncate_str(text='Se....',  max_len=4) == 'S...'
    assert truncate_str(text='Sen...',  max_len=4) == 'S...'
    assert truncate_str(text='Sent...', max_len=4) == 'S...'



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide :pep:`544`-compliant type hint utilities.

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from abc import abstractmethod
from beartype.typing import (
    Any,
    BinaryIO,
    Dict,
    IO,
    Optional,
    TextIO,
)
from beartype._data.api.standard.datatyping import TYPING_MODULE_NAMES
from beartype._data.cls.datacls import TYPES_PEP484_GENERIC_IO
from beartype._data.typing.datatypingport import (
    Hint,
    TypeIs,
)
from beartype._util.cls.utilclstest import is_type_builtin_or_fake
from typing import Protocol as typing_Protocol  # <-- unoptimized protocol

# ....................{ MAPPINGS                           }....................
# Initialized by the init_HINT_PEP484_IO_GENERIC_TO_PEP544_PROTOCOL() function.
HINT_PEP484_IO_GENERIC_TO_PEP544_PROTOCOL: Dict[type, Any] = {}
'''
Dictionary mapping from each :mod:`typing` **IO generic base class** (i.e.,
either :class:`typing.IO` itself *or* a subclass of :class:`typing.IO` defined
by the :mod:`typing` module) to the associated :mod:`beartype` **IO protocol**
(i.e., either :class:`_Pep544IO` itself *or* a subclass of :class:`_Pep544IO`
defined by this submodule).
'''

# ....................{ TESTERS                            }....................
#FIXME: We'd strongly prefer to annotate this as returning
#"TypeIs[typing_Protocol]" rather than "TypeIs[type]". The former is
#considerably more fine-grained and thus broadly useful than the latter. Sadly,
#both "mypy" and "pyright" complain about this. Both are wrong, of course:
#    beartype/_util/hint/pep/proposal/pep544.py:43: error: Variable
#    "typing.Protocol" is not valid as a type  [valid-type]
#
#Nonsense! "typing.Protocol" is *LITERALLY* a type. It's a type, guys. Like most
#types, it's both a type hint *AND* a type. That's fine. Oh, well.
def is_hint_pep484_generic_io(hint: Hint) -> TypeIs[type]:
    '''
    :data:`True` only if the passed object is a functionally useless
    :pep:`484`-compliant :mod:`typing` **IO generic superclass** (i.e., either
    :class:`typing.IO` itself *or* a subclass of :class:`typing.IO` defined by
    the :mod:`typing` module effectively unusable at runtime due to botched
    implementation details) that is losslessly replaceable with a useful
    :pep:`544`-compliant :mod:`beartype` **IO protocol** (i.e., either
    :class:`_Pep544IO` itself *or* a subclass of that class defined by this
    submodule intentionally designed to be usable at runtime).

    This tester is intentionally *not* memoized (e.g., by the
    :func:`callable_cached` decorator), as the implementation trivially reduces
    to an efficient one-liner.

    Parameters
    ----------
    hint : object
        Object to be inspected.

    Returns
    -------
    bool
        :data:`True` only if this object is a :pep:`484`-compliant IO generic
        base class.

    See Also
    --------
    :class:`_Pep544IO`
        Further commentary.
    '''

    # Avoid circular import dependencies.
    from beartype._util.hint.pep.utilpepget import (
        get_hint_pep_origin_type_or_none)

    # Type originating this hint, defined as Either:
    # * If this hint defines the "__origin__" dunder attribute...
    #   * Whose value is a type, that type (which is then said to "originate"
    #     this hint).
    #   * Whose value is *not* a type but this hint is a type, this hint itself
    #     (which is then said to "originate" itself).
    # * In all other cases, "None".
    hint_origin = get_hint_pep_origin_type_or_none(
        hint=hint, is_self_fallback=True)

    # Return true only if this originating type is either:
    # * An unsubscripted PEP 484-compliant IO generic base class
    #   (e.g., "typing.IO") *OR*....
    # * A subscripted PEP 484-compliant IO generic base class
    #   (e.g., "typing.IO[str]").
    return hint_origin in TYPES_PEP484_GENERIC_IO


#FIXME: We'd strongly prefer to annotate this as returning
#"TypeIs[typing_Protocol]" rather than "TypeIs[type]". See above for commentary.
def is_hint_pep544_protocol(hint: Hint) -> TypeIs[type]:
    '''
    :data:`True` only if the passed object is a :pep:`544`-compliant
    **protocol** (i.e., subclass of the :class:`typing.Protocol` superclass).

    This tester is intentionally *not* memoized (e.g., by the
    :func:`callable_cached` decorator), as the implementation trivially reduces
    to an efficient one-liner.

    Parameters
    ----------
    hint : object
        Object to be inspected.

    Returns
    -------
    bool
        :data:`True` only if this object is a :pep:`544`-compliant protocol.
    '''

    # Return true only if this hint is...
    return (
        # A type *AND*...
        isinstance(hint, type) and
        # A PEP 544-compliant protocol *AND*...
        issubclass(hint, typing_Protocol) and  # type: ignore[arg-type]
        # *NOT* a builtin type. For unknown reasons, some but *NOT* all
        # builtin types erroneously present themselves to be PEP
        # 544-compliant protocols under Python >= 3.8: e.g.,
        #     >>> from typing import Protocol
        #     >>> issubclass(str, Protocol)
        #     False        # <--- this makes sense
        #     >>> issubclass(int, Protocol)
        #     True         # <--- this makes no sense whatsoever
        #
        # Since builtin types are obviously *NOT* PEP 544-compliant
        # protocols, explicitly exclude all such types. Why, Guido? Why?
        #
        # Do *NOT* ignore fake builtins for the purposes of this test. Why?
        # Because even fake builtins (e.g., "type(None)") erroneously
        # masquerade as PEP 544-compliant protocols! :o
        not is_type_builtin_or_fake(hint)  # pyright: ignore
    )


#FIXME: Unit test us up, please.
def is_hint_pep544_protocol_supertype(hint: Hint) -> TypeIs[type]:
    '''
    :data:`True` only if the passed object is a :pep:`544`-compliant
    **protocol superclass** (i.e., either the :class:`typing.Protocol`,
    :class:`typing_extensions.Protocol`, or :class:`beartype.typing.Protocol`
    superclass).

    This tester is intentionally *not* memoized (e.g., by the
    :func:`callable_cached` decorator), as the implementation trivially reduces
    to an efficient one-liner.

    Parameters
    ----------
    hint : object
        Object to be inspected.

    Returns
    -------
    bool
        :data:`True` only if this object is a :pep:`544`-compliant protocol
        superclass.
    '''

    # Return true only if...
    return (
        # This object is a type *AND*...
        isinstance(hint, type) and
        # The unqualified basename of this type is that of *ALL* protocol
        # superclasses *AND*...
        hint.__name__ == 'Protocol' and
        # A typing-like module declares this type.
        hint.__module__ in TYPING_MODULE_NAMES
    )

# ....................{ FACTORIES                          }....................
def init_HINT_PEP484_IO_GENERIC_TO_PEP544_PROTOCOL() -> None:
    '''
    Initialize this submodule.
    '''

    # ..................{ IMPORTS                            }..................
    # Defer function-specific imports.
    from beartype.typing import (
        Annotated,
        AnyStr,
        List,
        Protocol,
    )
    from beartype.vale import IsInstance

    # ..................{ GLOBALS                            }..................
    # Global attributes to be redefined below.
    global \
        _Pep544BinaryIO, \
        _Pep544IO, \
        _Pep544TextIO

    # ..................{ PROTOCOLS ~ protocol               }..................
    #FIXME: Declare these protocols at global scope, please.
    # PEP-compliant type hint matching file handles opened in either text or
    # binary mode.
    # @runtime_checkable
    class _Pep544IO(Protocol[AnyStr]):
        # The body of this class is copied wholesale from the existing
        # non-functional "typing.IO" class.

        __slots__: tuple = ()

        @property
        @abstractmethod
        def mode(self) -> str:
            pass

        @property
        @abstractmethod
        def name(self) -> str:
            pass

        @abstractmethod
        def close(self) -> None:
            pass

        @property
        @abstractmethod
        def closed(self) -> bool:
            pass

        @abstractmethod
        def fileno(self) -> int:
            pass

        @abstractmethod
        def flush(self) -> None:
            pass

        @abstractmethod
        def isatty(self) -> bool:
            pass

        @abstractmethod
        def read(self, n: int = -1) -> AnyStr:
            pass

        @abstractmethod
        def readable(self) -> bool:
            pass

        @abstractmethod
        def readline(self, limit: int = -1) -> AnyStr:
            pass

        @abstractmethod
        def readlines(self, hint: int = -1) -> List[AnyStr]:
            pass

        @abstractmethod
        def seek(self, offset: int, whence: int = 0) -> int:
            pass

        @abstractmethod
        def seekable(self) -> bool:
            pass

        @abstractmethod
        def tell(self) -> int:
            pass

        @abstractmethod
        def truncate(self, size: Optional[int] = None) -> int:
            pass

        @abstractmethod
        def writable(self) -> bool:
            pass

        @abstractmethod
        def write(self, s: AnyStr) -> int:
            pass

        @abstractmethod
        def writelines(self, lines: List[AnyStr]) -> None:
            pass

        @abstractmethod
        def __enter__(self) -> '_Pep544IO[AnyStr]':  # pyright: ignore
            pass

        @abstractmethod
        def __exit__(self, cls, value, traceback) -> None:
            pass


    # PEP-compliant type hint matching file handles opened in text rather than
    # binary mode.
    #
    # Note that PEP 544 explicitly requires *ALL* protocols (including
    # protocols subclassing protocols) to explicitly subclass the "Protocol"
    # superclass, in violation of both sanity and usability. (Thanks, guys.)
    # @runtime_checkable
    class _Pep544TextIO(_Pep544IO[str], Protocol):
        # The body of this class is copied wholesale from the existing
        # non-functional "typing.TextIO" class.

        __slots__: tuple = ()

        @property
        @abstractmethod
        def buffer(self) -> _Pep544BinaryIO:  # pyright: ignore
            pass

        @property
        @abstractmethod
        def encoding(self) -> str:
            pass

        @property
        @abstractmethod
        def errors(self) -> Optional[str]:
            pass

        @property
        @abstractmethod
        def line_buffering(self) -> bool:
            pass

        @property
        @abstractmethod
        def newlines(self) -> Any:
            pass

        @abstractmethod
        def __enter__(self) -> '_Pep544TextIO':  # pyright: ignore
            pass

    # ..................{ PROTOCOLS ~ validator              }..................
    # PEP-compliant type hint matching file handles opened in binary rather
    # than text mode. Specifically, this hint matches the abstract "typing.IO"
    # protocol ABC but *NOT* the concrete "typing.TextIO" subprotocol
    # subclassing that ABC. Whereas the concrete "typing.TextIO" subprotocol
    # unambiguously matches *ONLY* file handles opened in text mode, the
    # concrete "typing.BinaryIO" subprotocol ambiguously matches file handles
    # opened in both text *AND* binary mode. As the following hypothetical
    # "_Pep544BinaryIO" subclass demonstrates, the "typing.IO" and
    # "typing.BinaryIO" APIs are identical except for method annotations:
    #     class _Pep544BinaryIO(_Pep544IO[bytes], Protocol):
    #         # The body of this class is copied wholesale from the existing
    #         # non-functional "typing.BinaryIO" class.
    #
    #         __slots__: tuple = ()
    #
    #         @abstractmethod
    #         def write(self, s: Union[bytes, bytearray]) -> int:
    #             pass
    #
    #         @abstractmethod
    #         def __enter__(self) -> '_Pep544BinaryIO':
    #             pass
    #
    # Sadly, the method annotations that differ between these APIs are
    # insufficient to disambiguate file handles at runtime. Why? Because most
    # file handles are C-based and thus lack *ANY* annotations whatsoever. With
    # respect to C-based file handles, these APIs are therefore identical.
    # Ergo, the "typing.BinaryIO" subprotocol is mostly useless at runtime.
    #
    # Note, however, that file handles are necessarily *ALWAYS* opened in either
    # text or binary mode. This strict dichotomy implies that any file handle
    # (i.e., object matching the "typing.IO" protocol) *NOT* opened in text mode
    # (i.e., not matching the "typing.TextIO" protocol) must necessarily be
    # opened in binary mode instead.
    _Pep544BinaryIO = Annotated[_Pep544IO, ~IsInstance[_Pep544TextIO]]

    # ..................{ MAPPINGS                           }..................
    # Dictionary mapping from each "typing" IO generic base class to the
    # associated IO protocol defined above.
    #
    # Note this global is intentionally modified in-place rather than
    # reassigned to a new dictionary. Why? Because the higher-level
    # reduce_hint_pep484_generic_io_to_pep544_protocol() function calling this
    # lower-level initializer has already imported this global.
    HINT_PEP484_IO_GENERIC_TO_PEP544_PROTOCOL.update({
        # Unsubscripted mappings.
        IO:        _Pep544IO,
        BinaryIO:  _Pep544BinaryIO,
        TextIO:    _Pep544TextIO,

        # Subscripted mappings, leveraging the useful observation that these
        # classes all self-cache by design: e.g.,
        #     >>> import typing
        #     >>> typing.IO[str] is typing.IO[str]
        #     True
        #
        # Note that we intentionally map:
        # * "IO[Any]" to the unsubscripted "_Pep544IO" rather than the
        #   subscripted "_Pep544IO[Any]". Although the two are semantically
        #   equivalent, the latter is marginally more space- and time-efficient
        #   to generate code for and thus preferable.
        # * "IO[bytes]" to the unsubscripted "_Pep544Binary" rather than the
        #   subscripted "_Pep544IO[bytes]". Why? Because the former applies
        #   meaningful runtime constraints, whereas the latter does *NOT*.
        # * "IO[str]" to the unsubscripted "_Pep544Text" rather than the
        #   subscripted "_Pep544IO[str]" -- for the same reason.
        #
        # Note that we intentionally avoid mapping parametrizations of "IO" by
        # type variables. Since there exist a countably infinite number of
        # such parametrizations, the parent
        # reduce_hint_pep484_generic_io_to_pep544_protocol() function calling
        # this function handles such parametrizations mostly intelligently.
        IO[Any]:   _Pep544IO,
        IO[bytes]: _Pep544BinaryIO,
        IO[str]:   _Pep544TextIO,
    })

# ....................{ PRIVATE ~ classes                  }....................
# Conditionally initialized by the _init() function below.
_Pep544IO: Any = None  # type: ignore[assignment]
'''
:pep:`544`-compliant protocol base class for :class:`_Pep544TextIO` and
:class:`_Pep544BinaryIO`.

This is an abstract, generic version of the return of open().

NOTE: This does not distinguish between the different possible classes (text
vs. binary, read vs. write vs. read/write, append-only, unbuffered). The TextIO
and BinaryIO subclasses below capture the distinctions between text vs. binary,
which is pervasive in the interface; however we currently do not offer a way to
track the other distinctions in the type system.

Design
------
This base class intentionally duplicates the contents of the existing
:class:`typing.IO` generic base class by substituting the useless
:class:`typing.Generic` superclass of the latter with the useful
:class:`typing.Protocol` superclass of the former. Why? Because *no* stdlib
classes excluding those defined by the :mod:`typing` module itself subclass
:class:`typing.IO`. However, :class:`typing.IO` leverages neither the
:class:`abc.ABCMeta` metaclass *nor* the :class:`typing.Protocol` superclass
needed to support structural subtyping. Therefore, *no* stdlib objects
(including those returned by the :func:`open` builtin) satisfy either
:class:`typing.IO` itself or any subclasses of :class:`typing.IO` (e.g.,
:class:`typing.BinaryIO`, :class:`typing.TextIO`). Therefore,
:class:`typing.IO` and all subclasses thereof are functionally useless for all
practical intents. The conventional excuse `given by Python maintainers to
justify this abhorrent nonsensicality is as follows <typeshed_>`__:

    There are a lot of "file-like" classes, and the typing IO classes are meant
    as "protocols" for general files, but they cannot actually be protocols
    because the file protocol isn't very well defined—there are lots of methods
    that exist on some but not all filelike classes.

Like most :mod:`typing`-oriented confabulation, that, of course, is bollocks.
Refactoring the family of :mod:`typing` IO classes from inveterate generics
into pragmatic protocols is both technically trivial and semantically useful,
because that is exactly what :mod:`beartype` does. It works. It necessitates
modifying three lines of existing code. It preserves backward compatibility. In
short, it should have been done a decade ago. If the file protocol "isn't very
well defined," the solution is to define that protocol with a rigorous type
hierarchy satisfying all possible edge cases. The solution is *not* to pretend
that no solutions exist, that the existing non-solution suffices, and instead
do nothing. Welcome to :mod:`typing`, where no one cares that nothing works as
advertised (or at all)... *and no one ever will.*

.. _typeshed:
   https://github.com/python/typeshed/issues/3225#issuecomment-529277448
'''


# Conditionally initialized by the _init() function below.
_Pep544BinaryIO: Any = None  # type: ignore[assignment]
'''
Typed version of the return of :func:`open` in binary mode.
'''


# Conditionally initialized by the _init() function below.
_Pep544TextIO: Any = None  # type: ignore[assignment]
'''
Typed version of the return of :func:`open` in text mode.
'''



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide **frozen dictionary class hierarchy** (i.e., private classes
implementing immutable mappings, preserving :math:`O(1)` complexity while
prohibiting modification).
'''

# ....................{ IMPORTS                            }....................
from beartype.roar import BeartypeKindFrozenDictException
from beartype.typing import (
    NoReturn,
    SupportsIndex,
    Tuple,
)
from beartype._util.utilobject import get_object_type_basename
from collections.abc import Mapping

# ....................{ CLASSES                            }....................
#FIXME: Consider attempting to generalize with type variables: e.g.,
#    class FrozenDict(dict[S, T]):
#In theory, that should help static type-checkers. In practice, they'll probably
#just complain about everything and be even more of a pain. We sigh. *sigh*
class FrozenDict(dict):
    '''
    **Frozen dictionary** (i.e., immutable mapping preserving :math:`O(1)`
    complexity while prohibiting modification).

    Instances of this dictionary are safely hashable and thus suitable for
    passing as parameters to memoized callables and classes (e.g., our core
    :class:`beartype.BeartypeConf` class).

    See Also
    --------
    https://stackoverflow.com/q/1151658/2809027
        StackOverflow question lightly inspiring this implementation.
    '''

    # ..................{ CLASS VARIABLES                    }..................
    # Slot all instance variables defined on this object to minimize the time
    # complexity of both reading and writing variables across frequently called
    # @beartype decorations. Slotting has been shown to reduce read and write
    # costs by approximately ~10%, which is non-trivial.
    __slots__ = ('_hash',)

    # ..................{ CLASS METHODS                      }..................
    @classmethod
    def fromkeys(cls, *args, **kwargs) -> 'FrozenDict':

        # Create and return a new immutable dictionary encapsulating the mutable
        # dictionary created and returned by the superclass method.
        #
        # Note that this implementation intentionally calls the dict.fromkeys()
        # class method directly rather than calling super().fromkey(). While
        # seemingly equivalent, the latter implicitly calls the __setitem__()
        # dunder method of this subclass, which then raises an exception.
        return cls(dict.fromkeys(*args, **kwargs))

    # ..................{ INITIALIZERS                       }..................
    def __init__(self, *args, **kwargs) -> None:

        # Instantiate this immutable dictionary with all passed parameters.
        super().__init__(*args, **kwargs)

        # Precompute the hash for this immutable dictionary at instantiation
        # time for both efficiency and safety.
        frozen_items = frozenset(self.items())  # <-- clever stuff
        self._hash = hash(frozen_items)  # <---- more clever stuff

    # ..................{ DUNDERS                            }..................
    def __hash__(self) -> int:  # type: ignore[override]
        '''
        Hash of all key-value pairs in this immutable dictionary.

        Defining this method satisfies the :class:`collections.abc.Hashable`
        abstract base class (ABC), enabling this dictionary to be used as in
        hashable containers (e.g., dictionaries, sets).
        '''

        # Return the precomputed hash for this immutable dictionary.
        return self._hash


    def __reduce_ex__(self, protocol: SupportsIndex) -> (
        Tuple[type, Tuple[dict]]):
        '''
        Pickle this immutable dictionary.

        Parameters
        ----------
        protocol : SupportsIndex
            Pickle protocol under which to pickle this immutable dictionary.

        Returns
        -------
        Tuple[type, Tuple[dict]]
            2-tuple suitable for pickling this immutable dictionary.
        '''

        # Dark magic is both dark and magical.
        return (type(self), (dict(self),))


    def __repr__(self) -> str:
        '''
        Machine-readable representation of this immutable dictionary.
        '''

        # Standard "dict"-based representation of the mutable dictionary
        # encapsulated by this immutable dictionary.
        dict_repr = super().__repr__()

        # Fully-qualified name of the possible subclass of this dictionary.
        type_name = get_object_type_basename(self)

        # Return an appropriate representation of this immutable dictionary.
        return f'{type_name}({dict_repr})'


    def __or__(self, other: Mapping) -> 'FrozenDict':
        '''
        Create and return a new immutable dictionary containing all key-value
        pairs contained in both the current and passed immutable dictionaries.

        Parameters
        ----------
        other: Mapping
            Possibly mutable dictionary to be added to this immutable
            dictionary.

        Returns
        -------
        FrozenDict
            Immutable dictionary adding the current and passed dictionaries.

        Raises
        ------
        BeartypeKindFrozenDictException
            If the passed dictionary is *not* actually a dictionary.
        '''

        # If the passed dictionary is *NOT* a dictionary, raise an exception.
        if not isinstance(other, Mapping):
            raise BeartypeKindFrozenDictException(
                f'Non-dictionary {repr(other)} not addable to '
                f'immutable dictionary {repr(self)}.'
            )
        # Else, the passed dictionary is a dictionary.

        # Standard dictionary uniting this and the passed dictionaries, defined
        # by trivially deferring to the builtin dict.__or__() dunder method.
        dict_united: dict = super().__or__(dict(other))  # type: ignore[misc]

        # Type of immutable dictionary to be created and returned.
        cls = type(self)

        # Create and return a new immutable dictionary wrapping this dictionary.
        return cls(dict_united)

    # ..................{ MUTATORS                           }..................
    # Override all mutators (i.e., "dict" methods attempting to modify the
    # current immutable dictionary) to raise exceptions instead.
    def __setitem__(self, key, value) -> NoReturn:
        raise BeartypeKindFrozenDictException(
            f'Immutable dictionary {repr(self)} '
            f'key {repr(key)} not settable to {repr(value)}.'
        )


    def __delitem__(self, key) -> NoReturn:
        raise BeartypeKindFrozenDictException(
            f'Immutable dictionary {repr(self)} '
            f'key {repr(key)} not deletable.'
        )


    def clear(self) -> NoReturn:
        raise BeartypeKindFrozenDictException(
            f'Immutable dictionary {repr(self)} not clearable.')


    def pop(self, key, default = None) -> NoReturn:
        raise BeartypeKindFrozenDictException(
            f'Immutable dictionary {repr(self)} '
            f'key {repr(key)} with default {repr(default)} not poppable.'
        )


    def popitem(self) -> NoReturn:
        raise BeartypeKindFrozenDictException(
            f'Immutable dictionary {repr(self)} not poppable.')


    def setdefault(self, key, default = None) -> NoReturn:
        raise BeartypeKindFrozenDictException(
            f'Immutable dictionary {repr(self)} '
            f'key {repr(key)} with default {repr(default)} not settable.'
        )


    def update(self, *args, **kwargs) -> NoReturn:
        raise BeartypeKindFrozenDictException(
            f'Immutable dictionary {repr(self)} '
            f'not updatable from positional arguments {repr(args)} '
            f'and keyword arguments {repr(kwargs)}.'
        )



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide **callable argument iterator utility** unit tests.

This submodule unit tests the public API of the private
:mod:`beartype._util.utilfunc.arg.utilfuncargiter` submodule.
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

# ....................{ TESTS ~ iterator                   }....................
def test_iter_func_args() -> None:
    '''
    Test the
    :func:`beartype._util.func.arg.utilfuncargtest.iter_func_args` generator.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype.roar._roarexc import _BeartypeUtilCallableException
    from beartype._util.func.arg.utilfuncargiter import (
        ArgKind,
        ArgMandatory,
        iter_func_args,
    )
    from beartype_test.a00_unit.data.func.data_func import (
        ClassOfMethods,
        func_args_0,
        func_args_1_flex_mandatory,
        func_args_1_varpos,
        func_args_1_kwonly_mandatory,
        func_args_2_flex_mandatory,
        func_args_2_kwonly_mixed,
        func_args_3_flex_mandatory_optional_varkw,
        func_args_5_flex_mandatory_varpos_kwonly_varkw,
    )
    from beartype_test.a00_unit.data.func.data_pep570 import (
        func_args_2_posonly_mixed,
        func_args_10_all_except_flex_mandatory,
    )
    from pytest import raises

    # ....................{ LOCALS                         }....................
    # Instance of an arbitrary class defining various kinds of methods, which
    # due to this instantiation have now been implicitly coerced into bound
    # method descriptors, which this iterator explicitly supports.
    object_of_methods = ClassOfMethods()

    # ....................{ PASS                           }....................
    # Assert this iterator returns the empty generator for an argumentless
    # callable, explicitly coerced into a tuple to trivialize testing.
    assert len(tuple(iter_func_args(func_args_0))) == 0

    # Assert this iterator returns the expected generator for argumentative
    # callables accepting multiple kinds of parameters.
    assert tuple(iter_func_args(func_args_1_flex_mandatory)) == (
        (ArgKind.POSITIONAL_OR_KEYWORD, 'had_one_fair_daughter', ArgMandatory,),
    )
    assert tuple(iter_func_args(func_args_1_varpos)) == (
        (ArgKind.VARIADIC_POSITIONAL, 'and_in_her_his_one_delight', ArgMandatory,),
    )
    assert tuple(iter_func_args(func_args_1_kwonly_mandatory)) == (
        (ArgKind.KEYWORD_ONLY, 'when_can_I_take_you_from_this_place', ArgMandatory,),
    )
    assert tuple(iter_func_args(func_args_2_flex_mandatory)) == (
        (ArgKind.POSITIONAL_OR_KEYWORD, 'thick_with_wet_woods', ArgMandatory,),
        (ArgKind.POSITIONAL_OR_KEYWORD, 'and_many_a_beast_therein', ArgMandatory,),
    )
    assert tuple(iter_func_args(func_args_2_kwonly_mixed)) == (
        (ArgKind.KEYWORD_ONLY, 'white_summer', 'So far I have gone to see you again.',),
        (ArgKind.KEYWORD_ONLY, 'hiding_your_face_in_the_palm_of_your_hands', ArgMandatory,),
    )
    assert tuple(iter_func_args(func_args_3_flex_mandatory_optional_varkw)) == (
        (ArgKind.POSITIONAL_OR_KEYWORD, 'and_the_wolf_tracks_her_there', ArgMandatory,),
        (ArgKind.POSITIONAL_OR_KEYWORD, 'how_hideously', "Its shapes are heap'd around!",),
        (ArgKind.VARIADIC_KEYWORD, 'rude_bare_and_high', ArgMandatory,),
    )
    assert tuple(iter_func_args(func_args_5_flex_mandatory_varpos_kwonly_varkw)) == (
        (ArgKind.POSITIONAL_OR_KEYWORD, 'we_are_selfish_men', ArgMandatory,),
        (ArgKind.POSITIONAL_OR_KEYWORD, 'oh_raise_us_up', ArgMandatory,),
        (ArgKind.VARIADIC_POSITIONAL, 'and_give_us', ArgMandatory,),
        (ArgKind.KEYWORD_ONLY, 'return_to_us_again', 'Of inward happiness.',),
        (ArgKind.VARIADIC_KEYWORD, 'manners_virtue_freedom_power', ArgMandatory,),
    )

    # Assert this iterator returns the expected generator for argumentative
    # bound method descriptors accepting multiple kinds of parameters. Note that
    # this generator intentionally omits the first "self" parameters passed by
    # these descriptors to the unbound methods they encapsulate.
    assert tuple(iter_func_args(
        object_of_methods.meth_args_6_flex_mandatory_varpos_kwonly_varkw)) == (
        (ArgKind.POSITIONAL_OR_KEYWORD, 'calm_and_rejoicing', ArgMandatory,),
        (ArgKind.POSITIONAL_OR_KEYWORD, 'in_the_fearful_war', ArgMandatory,),
        (ArgKind.VARIADIC_POSITIONAL, 'of_wave_ruining_on_wave', ArgMandatory,),
        (ArgKind.KEYWORD_ONLY, 'and_blast_on_blast', 'Descending,',),
        (ArgKind.VARIADIC_KEYWORD, 'and_black_flood', ArgMandatory,),
    )

    # ....................{ PASS ~ positional-only         }....................
    # Assert this iterator returns the expected generator for argumentative
    # callables accepting multiple kinds of parameters -- including
    # positional-only parameters.
    assert tuple(iter_func_args(func_args_2_posonly_mixed)) == (
        (ArgKind.POSITIONAL_ONLY, 'before_spreading_his_black_wings', ArgMandatory,),
        (ArgKind.POSITIONAL_ONLY, 'reaching_for_the_skies', 'in this forest',),
    )
    assert tuple(iter_func_args(func_args_10_all_except_flex_mandatory)) == (
        (ArgKind.POSITIONAL_ONLY, 'in_solitude_i_wander', ArgMandatory,),
        (ArgKind.POSITIONAL_ONLY, 'through_the_vast_enchanted_forest', ArgMandatory,),
        (ArgKind.POSITIONAL_ONLY, 'the_surrounding_skies', 'are one',),
        (ArgKind.POSITIONAL_OR_KEYWORD, 'torn_apart_by', 'the phenomenon of lightning',),
        (ArgKind.POSITIONAL_OR_KEYWORD, 'rain_is_pouring_down', 'my now shivering shoulders',),
        (ArgKind.VARIADIC_POSITIONAL, 'in_the_rain_my_tears_are_forever_lost', ArgMandatory,),
        (ArgKind.KEYWORD_ONLY, 'the_darkened_oaks_are_my_only_shelter', ArgMandatory,),
        (ArgKind.KEYWORD_ONLY, 'red_leaves_are_blown_by', 'the wind',),
        (ArgKind.KEYWORD_ONLY, 'an_ebony_raven_now_catches', 'my eye.',),
        (ArgKind.VARIADIC_KEYWORD, 'sitting_in_calmness', ArgMandatory,),
    )

    # ....................{ FAIL                           }....................
    # Assert this iterator returns a generator raising the expected exception
    # when passed a C-based callable.
    with raises(_BeartypeUtilCallableException):
        next(iter_func_args(iter))



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide **object factory utility** unit tests.

This submodule unit tests the public API of the private
:mod:`beartype._util.utilobjmake` submodule.
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

# ....................{ TESTS ~ getter                     }....................
def test_permute_object() -> None:
    '''
    Test both the :func:`beartype._util.utilobjmake.permute_object` factory.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype.roar._roarexc import _BeartypeUtilTypeException
    from beartype.typing import Optional
    from beartype._util.utilobjmake import permute_object
    from pytest import raises

    # ....................{ CLASSES                        }....................
    class FoundWayOntoOlympus(object):
        '''
        Arbitrary class.
        '''

        def __init__(
            self,

            # Mandatory parameters.
            and_made_quake: bytes,
            the_rebel_three: str,

            # Optional parameters.
            yet_full_of_awe: Optional[int] = None,
        ) -> None:
            '''
            Arbitrary constructor accepting two mandatory and one optional
            parameters.
            '''

            # Default this optional parameter if unpassed.
            if yet_full_of_awe is None:
                yet_full_of_awe = 0xFEEDFACE

            # Classify all passed parameters.
            self.and_made_quake = and_made_quake
            self.the_rebel_three = the_rebel_three
            self.yet_full_of_awe = yet_full_of_awe


        def __eq__(self, other: object) -> bool:
            '''
            Arbitrary equality tester defined in the trivial way.
            '''

            return (
                isinstance(other, FoundWayOntoOlympus) and
                self.and_made_quake == other.and_made_quake and
                self.the_rebel_three == other.the_rebel_three and
                self.yet_full_of_awe == other.yet_full_of_awe
            )

    # ....................{ LOCALS                         }....................
    # Frozen set of the names of all mandatory parameters accepted by the
    # FoundWayOntoOlympus.__init__() method defined above.
    COPY_VAR_NAMES = frozenset((
        'and_made_quake',
        'the_rebel_three',
    ))

    # Frozen set of the names of all mandatory *AND* optional parameters
    # accepted by the FoundWayOntoOlympus.__init__() method defined above.
    INIT_ARG_NAMES = COPY_VAR_NAMES | frozenset(('yet_full_of_awe',))

    # Arbitrary instance of this class.
    thea_was_startled_up = FoundWayOntoOlympus(
        and_made_quake=b'Found way unto Olympus, and made quake',
        the_rebel_three='The rebel three.—Thea was startled up,',
        yet_full_of_awe=0xCAFEBABE,
    )

    # ....................{ PASS                           }....................
    # Assert that passing this function an empty parameter dictionary returns a
    # shallow copy of the passed object.
    as_thus_she_quick = permute_object(
        obj=thea_was_startled_up,
        init_arg_name_to_value={},
        init_arg_names=INIT_ARG_NAMES,
    )
    assert as_thus_she_quick == thea_was_startled_up

    # Assert that passing this function a non-empty parameter dictionary setting
    # only one of the parameters accepted by the FoundWayOntoOlympus.__init__()
    # method returns a new "FoundWayOntoOlympus" object defaulting the unpassed
    # parameter to the instance variable of the same name of the passed object.
    as_thus_she_quick = permute_object(
        obj=thea_was_startled_up,
        init_arg_name_to_value={
            'and_made_quake': b'spake, yet full of awe.',},
        init_arg_names=INIT_ARG_NAMES,
    )
    assert as_thus_she_quick == FoundWayOntoOlympus(
        and_made_quake=b'spake, yet full of awe.',
        the_rebel_three='The rebel three.—Thea was startled up,',
        yet_full_of_awe=0xCAFEBABE,
    )

    # Assert that passing this function an empty parameter dictionary *AND* an
    # optional parameter names frozen set returns a new "FoundWayOntoOlympus"
    # object defaulting:
    # * All unpassed mandatory parameters to the instance variables of the
    #   same names of the passed object.
    # * The unpassed optional parameter to default value assigned to that
    #   parameter by the FoundWayOntoOlympus.__init__() method.
    as_thus_she_quick = permute_object(
        obj=thea_was_startled_up,
        init_arg_name_to_value={},
        init_arg_names=INIT_ARG_NAMES,
        copy_var_names=COPY_VAR_NAMES,
    )
    assert as_thus_she_quick == FoundWayOntoOlympus(
        and_made_quake=b'Found way unto Olympus, and made quake',
        the_rebel_three='The rebel three.—Thea was startled up,',
        yet_full_of_awe=0xFEEDFACE,
    )

    # ....................{ FAIL                           }....................
    # Assert that passing this function an parameter dictionary containing an
    # parameter *NOT* accepted by the FoundWayOntoOlympus.__init__() method
    # raises the expected exception.
    with raises(_BeartypeUtilTypeException):
        permute_object(
            obj=thea_was_startled_up,
            init_arg_name_to_value={
                'and_in_her_bearing': 'was a sort of hope',},
            init_arg_names=INIT_ARG_NAMES,
        )

    # Assert that passing this function an default parameter set containing an
    # parameter *NOT* accepted by the FoundWayOntoOlympus.__init__() method
    # raises the expected exception.
    with raises(_BeartypeUtilTypeException):
        permute_object(
            obj=thea_was_startled_up,
            init_arg_name_to_value={},
            init_arg_names=INIT_ARG_NAMES,
            copy_var_names=frozenset(('this_cheers_our_fallen_house',)),
        )



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Beartype PEP-noncompliant class type hint violation describers** (i.e.,
functions returning human-readable strings explaining violations of type hints
that are PEP-noncompliant isinstanceable classes rather than PEP-compliant type
hints).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype.roar import (
    BeartypeCallHintForwardRefException,
    BeartypePlugInstancecheckStrException,
)
from beartype.roar._roarexc import _BeartypeCallHintPepRaiseException
from beartype.typing import Optional
from beartype._check.error.errcause import ViolationCause
from beartype._data.typing.datatyping import TupleTypes
from beartype._data.hint.sign.datahintsigns import HintSignForwardRef
from beartype._util.cls.pep.clspep3119 import die_unless_type_isinstanceable
from beartype._util.func.arg.utilfuncargtest import (
    die_unless_func_args_len_flexible_equal)
from beartype._util.hint.nonpep.utilnonpeptest import (
    die_unless_hint_nonpep_tuple)
from beartype._util.hint.pep.proposal.pep484585.pep484585ref import (
    import_pep484585_ref_type)
from beartype._util.hint.pep.utilpepget import (
    get_hint_pep_origin_type_isinstanceable_or_none)
from beartype._util.text.utiltextjoin import join_delimited_disjunction_types
from beartype._util.text.utiltextlabel import label_type
from beartype._util.text.utiltextrepr import represent_pith

# ....................{ GETTERS ~ instance : type          }....................
def find_cause_nonpep(cause: ViolationCause) -> ViolationCause:
    '''
    Output cause describing whether the pith of the passed input cause either
    does or does not satisfy the PEP-noncompliant type hint of that cause.

    Parameters
    ----------
    cause : ViolationCause
        Input cause providing this data.

    Returns
    -------
    ViolationCause
        Output cause type-checking this data.
    '''
    assert isinstance(cause, ViolationCause), f'{repr(cause)} not cause.'

    # If this PEP-noncompliant hint is a tuple union, defer to the finder
    # specific to tuple unions.
    if isinstance(cause.hint, tuple):
        cause_finder = find_cause_instance_types_tuple
    # Else, this PEP-noncompliant hint is *NOT* a tuple union. In this case,
    # assume this hint to be an isinstanceable class by deferring to the finder
    # specific to isinstanceable classes. When this assumption is incorrect
    # (i.e., if this hint is *NOT* an isinstanceable class), this finder raises
    # a general-purpose human-readable exception.
    else:
        cause_finder = find_cause_instance_type

    # Trivially defer to this finder.
    return cause_finder(cause)

# ....................{ GETTERS ~ instance : type          }....................
def find_cause_instance_type(cause: ViolationCause) -> ViolationCause:
    '''
    Output cause describing whether the pith of the passed input cause either is
    or is not an instance of the isinstanceable class of that cause.

    Parameters
    ----------
    cause : ViolationCause
        Input cause providing this data.

    Returns
    -------
    ViolationCause
        Output cause type-checking this data.

    Raises
    ------
    BeartypePlugInstancecheckStrException
        If the metaclass of this isinstanceable class defines the
        :mod:`beartype`-specific ``__instancecheck_str__()`` dunder method but
        either:

        * This method is *not* a pure-Python callable.
        * This method is a pure-Python callable with an unexpected signature
          that differs from the expected API:

          .. code-block:: python

             def __instancecheck_str__(cls, obj: typing.Any) -> str:

        * This method is a pure-Python callable with the expected signature that
          returns either:

          * An object that is *not* a string.
          * The empty string.
    ViolationCause
        Output cause type-checking this data.
    '''
    assert isinstance(cause, ViolationCause), f'{repr(cause)} not cause.'

    # Isinstanceable class against which this pith was type-checked.
    hint: type = cause.hint  # type: ignore[assignment]

    # Pith type-checked against this isinstanceable class.
    pith = cause.pith

    # If this hint is *NOT* an isinstanceable class, raise an exception.
    die_unless_type_isinstanceable(
        cls=hint,
        exception_cls=_BeartypeCallHintPepRaiseException,
        exception_prefix=cause.exception_prefix,
    )
    # Else, this hint is an isinstanceable class.

    # Output cause justification. If this pith either:
    # * Violates this hint, this is a human-readable substring describing this
    #   violation.
    # * Satisfies this hint, "None".
    cause_str_or_none: Optional[str] = None

    # If this pith is *NOT* an instance of this class...
    if not isinstance(pith, hint):
        # Metaclass-specific __instancecheck_str__() dunder method if the
        # metaclass of this class defines this method *OR* "None" otherwise
        # (i.e., if that metaclass does *NOT* define this method).
        #
        # Note that this constitutes a plugin API. Although currently
        # beartype-specific, this API is intended to receive widespread adoption
        # as a pseudo-standard throughout the runtime type-checking community
        # (e.g., by typeguard and possibly Pydantic). Various third-party
        # packages that publish custom type hint factories currently leverage
        # this API to generate package-specific violation messages, including:
        # * @patrick-kidger's "jaxtyping" package. For the good of Google!
        get_hint_violation_str = getattr(hint, '__instancecheck_str__', None)

        # If the metaclass of this class defines this dunder method...
        if get_hint_violation_str:
            # Human-readable substring prefixing *ALL* exceptions raised below.
            EXCEPTION_PREFIX = (
                f'{cause.exception_prefix}{repr(hint)} '
                f'beartype-specific dunder method __instancecheck_str__() '
            )

            # If this method is *NOT* a pure-Python callable accepting exactly
            # two parameters, this method does *NOT* satisfy the expected API:
            #      def __instancecheck_str__(cls, obj: typing.Any) -> str:
            #
            # In this case, raise an exception.
            die_unless_func_args_len_flexible_equal(
                func=get_hint_violation_str,
                func_args_len_flexible=2,
                exception_cls=BeartypePlugInstancecheckStrException,
                exception_prefix=EXCEPTION_PREFIX,
            )
            # Else, this method satisfies the expected API.

            # Human-readable substring describing this violation generated by
            # the metaclass of this class.
            cause_str_or_none = get_hint_violation_str(pith)

            # If this string is *NOT* actually a string, raise an exception.
            if not isinstance(cause_str_or_none, str):
                raise BeartypePlugInstancecheckStrException(
                    f'{EXCEPTION_PREFIX}return {cause_str_or_none} not string.')
            # Else, this string is actually a string.
            #
            # If this string is empty, raise an exception.
            elif not cause_str_or_none:
                raise BeartypePlugInstancecheckStrException(
                    f'{EXCEPTION_PREFIX}return string empty.')
            # Else, this string is non-empty.
        # Else, the metaclass of this class does *NOT* define this method. In
        # this case, fallback to a standard substring describing this violation.
        else:
            cause_str_or_none = (
                f'{represent_pith(pith)} not instance of '
                f'{label_type(cls=hint, is_color=cause.conf.is_color)}'
            )
    # Else, this pith is an instance of this class.

    # Output cause to be returned, permuted from this input cause with this
    # output cause justification.
    cause_return = cause.permute_cause(cause_str_or_none=cause_str_or_none)

    # Return this output cause.
    return cause_return


def find_cause_instance_type_forwardref(
    cause: ViolationCause) -> ViolationCause:
    '''
    Output cause describing whether the pith of the passed input cause either is
    or is not an instance of the class referred to by the **forward reference
    type hint** (i.e., string whose value is the either absolute *or* relative
    name of a user-defined type which has yet to be defined) of that cause.

    Parameters
    ----------
    cause : ViolationCause
        Input cause providing this data.

    Returns
    -------
    ViolationCause
        Output cause type-checking this data.
    '''
    assert isinstance(cause, ViolationCause), f'{repr(cause)} not cause.'
    assert cause.hint_sign is HintSignForwardRef, (
        f'{cause.hint_sign} not forward reference.')

    # Class referred to by this absolute or relative forward reference.
    hint_ref_type = import_pep484585_ref_type(
        hint=cause.hint,  # type: ignore[arg-type]
        cls_stack=cause.cls_stack,
        func=cause.func,
        exception_cls=BeartypeCallHintForwardRefException,
        exception_prefix=cause.exception_prefix,
    )

    # Output cause to be returned.
    cause_return = cause.permute_cause_hint_child_insane(hint_ref_type)

    # Defer to the function handling isinstanceable classes. Neato!
    return find_cause_instance_type(cause_return)


def find_cause_type_instance_origin(cause: ViolationCause) -> ViolationCause:
    '''
    Output cause describing whether the pith of the passed input cause either is
    or is not an instance of the isinstanceable type underlying the
    **originative type hint** (i.e., PEP-compliant type hint originating from a
    non-:mod:`typing` class, typically due to being either a
    :pep:`585`-compliant type hint *or* a third-party type hint subclassing the
    :class:`types.GenericAlias` superclass defined by :pep:`585`) of that cause.

    Parameters
    ----------
    cause : ViolationCause
        Input cause providing this data.

    Returns
    -------
    ViolationCause
        Output cause type-checking this data.
    '''
    assert isinstance(cause, ViolationCause), f'{repr(cause)} not cause.'

    # Isinstanceable origin type originating this hint if any *OR* "None".
    hint_type = get_hint_pep_origin_type_isinstanceable_or_none(cause.hint)

    # If this hint does *NOT* originate from such a type, raise an exception.
    if hint_type is None:
        raise _BeartypeCallHintPepRaiseException(
            f'{cause.exception_prefix}type hint '
            f'{repr(cause.hint)} not originated from '
            f'isinstanceable origin type.'
        )
    # Else, this hint originates from such a type.

    # Output cause to be returned.
    cause_return = cause.permute_cause_hint_child_insane(hint_type)

    # Defer to the getter function handling non-"typing" classes. Presto!
    return find_cause_instance_type(cause_return)

# ....................{ GETTERS ~ instance : types         }....................
def find_cause_instance_types_tuple(cause: ViolationCause) -> ViolationCause:
    '''
    Output cause describing whether the pith of the passed input cause either is
    or is not an instance of one or more isinstanceable types in the tuple of
    these types of that cause.

    Parameters
    ----------
    cause : ViolationCause
        Input cause providing this data.

    Returns
    -------
    ViolationCause
        Output cause type-checking this data.
    '''
    assert isinstance(cause, ViolationCause), f'{repr(cause)} not cause.'

    # This tuple union.
    hint: TupleTypes = cause.hint  # type: ignore[assignment]

    # If this hint is *NOT* a tuple union, raise an exception.
    die_unless_hint_nonpep_tuple(
        hint=hint,
        exception_prefix=cause.exception_prefix,
        exception_cls=_BeartypeCallHintPepRaiseException,
    )
    # Else, this hint is a tuple union.

    # If this pith is an instance of one or more types in this tuple union,
    # record that this pith satisfies this tuple union.
    if isinstance(cause.pith, hint):
        cause_return = cause.permute_cause(cause_str_or_none=None)
    # Else, this pith is an instance of *NO* types in this tuple union. In
    # this case, this pith violates this tuple union.
    else:
        # Machine-readable representation of this tuple union.
        hint_repr = join_delimited_disjunction_types(
            types=hint, is_color=cause.conf.is_color)

        # Output cause to be returned, permuted from this input cause such that
        # the output cause justification is a substring describing this failure.
        cause_return = cause.permute_cause(cause_str_or_none=(
            f'{represent_pith(cause.pith)} not instance of {hint_repr}'))

    # Return this output cause.
    return cause_return



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Beartype **abstract syntax tree (AST) testers** (i.e., low-level callables
testing various properties of various nodes in the currently visited AST).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype._data.typing.datatyping import NodeCallable

# ....................{ TESTERS                            }....................
def is_node_callable_typed(node: NodeCallable) -> bool:
    '''
    :data:`True` only if the passed **callable node** (i.e., node signifying the
    definition of a pure-Python function or method) is **typed** (i.e.,
    annotated by a return type hint and/or one or more parameter type hints).

    Parameters
    ----------
    node : NodeCallable
        Callable node to be tested.

    Returns
    -------
    bool
        :data:`True` only if this callable node is typed.
    '''

    # Note that this algorithm is intentionally implemented in an unintuitive
    # order so as to increase the likelihood of efficiently deciding this
    # problem in best-case O(1) time. Specifically:
    # * It is most efficient to test whether that callable is annotated by a
    #   return type hint.
    # * It is next-most efficient to test whether that callable accepts a
    #   variadic positional or keyword argument annotated by a type hint.
    # * It is least efficient to test whether that callable accepts a
    #   non-variadic parameter annotated by a type hint, as doing so requires
    #   O(n) iteration for "n" the number of such arguments.
    #
    # Lastly, note that we could naively avoid doing this entirely and instead
    # unconditionally decorate *ALL* callables by @beartype -- in which case
    # @beartype would simply reduce to a noop for untyped callables annotated by
    # *NO* type hints. Technically, that works. Pragmatically, that would almost
    # certainly be slower than the current approach under the common assumption
    # that any developer annotating one or more non-variadic parameters of a
    # callable would also annotate the return of that callable -- in which case
    # this detection reduces to O(1) time complexity. Even where this is *NOT*
    # the case, however, this is still almost certainly slightly faster or of an
    # equivalent speed to the naive approach. Why? Because treating untyped
    # callables as typed would needlessly:
    # * Increase space complexity by polluting this AST with needlessly many
    #   "Name" child nodes performing untyped @beartype decorations.
    # * Increase time complexity by instantiating, initializing, and inserting
    #   (the three dread i's) those nodes.
    #
    # Admittedly, this approach is *CONSIDERABLY* slower for untyped callables,
    # where this detection exhibits worst-case O(n) time complexity. In theory,
    # the "beartype.claw" API that calls this tester function once per callable
    # should *NEVER* be applied to untyped callables. In practice, that API
    # almost certainly will be. We (largely) consider that the responsibility of
    # the caller, however. Beartype can't be faulted for optimizing for the
    # ideal case of well-typed packages... *CAN IT*!?!? o_O

    # If the return of that callable is typed, that callable is typed. In this
    # case, immediately and efficiently return true.
    if node.returns:
        return True
    # Else, the return of that callable is untyped.

    # Child arguments node of all arguments accepted by that callable.
    node_arg_nodes = node.args

    # If either...
    #
    # Note that PEP 484-compliant typed variadic positional arguments (e.g.,
    # "*args: str") are considerably more common than PEP 692-compliant typed
    # variadic keyword arguments (e.g., "**kwargs: SomeTypedDict", where
    # "SomeTypedDict" is a user-defined "typing.TypedDict" subclass). Ergo, we
    # intentionally detect typed variadic positional arguments *BEFORE* typed
    # variadic keyword arguments.
    if (
        (
            # That callable accepts a variadic positional argument *AND*...
            node_arg_nodes.vararg and
            # That parameter is typed...
            node_arg_nodes.vararg.annotation
        # *OR*...
        ) or
        (
            # That callable accepts a variadic keyword argument *AND*...
            node_arg_nodes.kwarg and
            # That parameter is typed...
            node_arg_nodes.kwarg.annotation
        )
    # That callable is typed. In this case, return true.
    ):
        return True
    # Else, that callable is still possibly untyped.

    # Fallback to deciding whether that callable accepts one or more typed
    # non-variadic parameters. Since doing is considerably more computationally
    # expensive, we do so *ONLY* as needed.
    #
    # Note that manual iteration is considerably more efficient than more
    # syntactically concise any() and all() generator expressions.
    #
    # Specifically, if that callable accepts non-variadic flexible parameters...
    if node_arg_nodes.args:
        # For each non-variadic flexible parameter...
        for node_arg_nonvar in node_arg_nodes.args:
            # If this parameter is typed, that callable is typed. In this case,
            # return true.
            if node_arg_nonvar.annotation:
                return True
            # Else, this parameter is untyped. Continue to the next.
    # Else, that callable is still possibly untyped.

    # If that callable accepts non-variadic keyword-only parameters...
    if node_arg_nodes.kwonlyargs:
        # For each non-variadic keyword-only parameter...
        for node_arg_nonvar in node_arg_nodes.kwonlyargs:
            # If this parameter is typed, that callable is typed. In this case,
            # return true.
            if node_arg_nonvar.annotation:
                return True
            # Else, this parameter is untyped. Continue to the next.
    # Else, that callable is still possibly untyped.

    # If that callable accepts non-variadic positional-only parameters...
    if node_arg_nodes.posonlyargs:
        # For each non-variadic positional-only parameter...
        for node_arg_nonvar in node_arg_nodes.posonlyargs:
            # If this parameter is typed, that callable is typed. In this case,
            # return true.
            if node_arg_nonvar.annotation:
                return True
            # Else, this parameter is untyped. Continue to the next.
    # Else, that callable is now known to be untyped.

    # Return false.
    return False



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide :pep:`695`-compliant **type alias reducers** (i.e., low-level
callables converting higher-level objects created via the ``type`` statement
under Python >= 3.12 to lower-level type hints more readily consumable by
:mod:`beartype`).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype.roar import BeartypeDecorHintPep695Exception
from beartype.typing import Optional
from beartype._cave._cavefast import HintPep695TypeAlias
from beartype._check.convert._reduce._redrecurse import (
    is_hint_recursive,
    make_hint_sane_recursable,
)
from beartype._check.metadata.hint.hintsane import (
    HINT_SANE_RECURSIVE,
    HintOrSane,
    HintSane,
)
from beartype._data.typing.datatypingport import Hint
from beartype._util.error.utilerrget import get_name_error_attr_name
from beartype._util.hint.pep.proposal.pep695 import (
    get_hint_pep695_unsubbed_alias)

# ....................{ REDUCERS                           }....................
def reduce_hint_pep695_subbed(
    hint: Hint,
    hint_parent_sane: Optional[HintSane],
    exception_prefix: str,
    **kwargs,
) -> HintOrSane:
    '''
    Reduce the passed :pep:`695`-compliant **subscripted type alias** (i.e.,
    object created by a statement of the form ``type
    {alias_name}[{typevar_name}] = {alias_value}``) to that unsubscripted alias
    and corresponding **type variable lookup table** (i.e., immutable dictionary
    mapping from those same type variables to those same child hints).

    This reducer is intentionally *not* memoized (e.g., by the
    ``callable_cached`` decorator), as reducers cannot be memoized.

    Parameters
    ----------
    hint : Hint
        Subscripted hint to be inspected.
    hint_parent_sane : Optional[HintSane]
        Either:

        * If the passed hint is a **root** (i.e., top-most parent hint of a tree
          of child hints), :data:`None`.
        * Else, the passed hint is a **child** of some parent hint. In this
          case, the **sanified parent type hint metadata** (i.e., immutable and
          thus hashable object encapsulating *all* metadata previously returned
          by :mod:`beartype._check.convert.convmain` sanifiers after
          sanitizing the possibly PEP-noncompliant parent hint of this child
          hint into a fully PEP-compliant parent hint).
    exception_prefix : str
        Human-readable substring prefixing raised exception messages.

    Returns
    -------
    HintSane
        **Sanified type hint metadata** (i.e., :class:`.HintSane` object)
        describing this reduction.

    Raises
    ------
    BeartypeDecorHintPep484TypeVarViolation
        If one of these type hints violates the bounds or constraints of one of
        these type variables.

    See Also
    --------
    ``reduce_hint_pep484612646_subbed_typeargs_to_hints``
        Further details.
    '''
    # print(f'Reducing PEP 695 subscripted type alias {hint} with parent {hint_parent_sane}...')

    # ....................{ IMPORTS                        }....................
    # Avoid circular import dependencies.
    from beartype._check.convert._reduce._pep.redpep484612646 import (
        reduce_hint_pep484612646_subbed_typeargs_to_hints)

    # ....................{ RECURSE                        }....................
    # If this PEP 695-compliant subscripted type alias is recursive, ignore this
    # recursive alias to avoid infinite recursion.
    #
    # Certainly, various approaches to generating code type-checking recursive
    # hints exists. @beartype currently embraces the easiest, fastest, and
    # laziest approach: just ignore all recursion! Ignorance works wonders.
    if is_hint_recursive(
        hint=hint,
        hint_parent_sane=hint_parent_sane,
        hint_recursable_depth_max=_HINT_PEP695_RECURSABLE_DEPTH_MAX,
    ):
        # print(f'Ignoring recursive PEP 695 subscripted type alias {hint} with parent {hint_parent_sane}...')
        return HINT_SANE_RECURSIVE
    # Else, this hint is *NOT* recursive.

    # ....................{ PHASE                          }....................
    # This reducer is divided into two phases:
    # 1. The first phase decides the type variable lookup table for this alias.
    # 2. The second phase decides the recursion guard for this alias.
    #
    # Both phases are non-trivial. The output of each phase is sanified hint
    # metadata (i.e., a "HintSane" object) containing the result of the decision
    # problem decided by that phase.

    # ....................{ PHASE ~ 1                      }....................
    # Decide the type variable lookup table for this alias. Specifically, reduce
    # this PEP 695-compliant subscripted type alias to:
    # * The semantically useful unsubscripted alias originating this
    #   semantically useless subscripted alias.
    # * The type variable lookup table mapping all type variables parametrizing
    #   this unsubscripted alias to all non-type variable hints subscripting
    #   this subscripted alias.
    # print(f'[reduce_hint_pep484585_generic_subbed] Reducing subscripted generic {repr(hint)}...')
    hint_or_sane = reduce_hint_pep484612646_subbed_typeargs_to_hints(
        hint=hint,
        hint_parent_sane=hint_parent_sane,
        exception_prefix=exception_prefix,
    )

    # ....................{ PHASE ~ 2                      }....................
    # If the prior phase generated metadata...
    if isinstance(hint_or_sane, HintSane):
        # Non-recursable form of this type alias, defined as the *UNSUBSCRIPTED*
        # type alias encapsulated by this metadata.
        hint_nonrecursable = hint_or_sane.hint

        # Sanified parent type hint metadata encapsulating the sanification of
        # both the parent hint (if any) *AND* the previously decided type
        # variable lookup table for this alias. Since this new metadata is
        # guaranteed to be the superset of the old metadata applying *ONLY* to
        # the parent hint, we intentionally replace the latter with the former
        # here. See also further discussion below.
        hint_parent_sane = hint_or_sane
    # Else, the prior phase generated *NO* metadata. In this case...
    else:
        # Non-recursable form of this type alias, defined as the *UNSUBSCRIPTED*
        # type alias directly returned by the prior call to the
        # reduce_hint_pep484612646_subbed_typeargs_to_hints() reducer.
        hint_nonrecursable = hint_or_sane

    # Decide the recursion guard protecting this possibly recursive alias
    # against infinite recursion. Note that:
    # * This guard intentionally applies to the original *SUBSCRIPTED* PEP
    #   695-compliant type alias (rather rather than the *UNSUBSCRIPTED* PEP
    #   695-compliant type alias decided by the prior phase). Thus, we pass
    #   "hint_recursable=hint" rather than "hint_recursable=hint_or_sane.hint".
    # * The type variable lookup table decided in the first phase *MUST* also be
    #   preserved. Thus, we pass a new "hint_parent_sane" rather than the same
    #   "hint_parent_sane" as in the prior phase. Indeed, the new
    #   "hint_parent_sane" object should safely encapsulate all metadata
    #   encapsulated by the prior "hint_parent_sane" object.
    hint_sane = make_hint_sane_recursable(
        # The recursable form of this type alias is the original *SUBSCRIPTED*
        # type alias tested above by the is_hint_recursive() recursion guard.
        hint_recursable=hint,
        # The non-recursable form of this type alias is the new *UNSUBSCRIPTED*
        # type alias encapsulated by the metadata returned by the prior call to
        # the reduce_hint_pep484612646_subbed_typeargs_to_hints() reducer.
        hint_nonrecursable=hint_nonrecursable,
        hint_parent_sane=hint_parent_sane,
    )

    # ....................{ RETURN                         }....................
    # Return this metadata.
    return hint_sane


def reduce_hint_pep695_unsubbed(
    hint: Hint,
    hint_parent_sane: Optional[HintSane],
    exception_prefix: str,
    **kwargs,
) -> HintOrSane:
    '''
    Reduce the passed :pep:`695`-compliant **unsubscripted type alias** (i.e.,
    object created by a statement of the form ``type {alias_name} =
    {alias_value}``) to the underlying type hint referred to by this alias.

    This reducer is intentionally *not* memoized (e.g., by the
    ``callable_cached`` decorator), as this reducer accepts the contextual
    ``hint_parent_sane`` parameter and thus *cannot* be memoized.

    Parameters
    ----------
    hint : HintPep695TypeAlias
        Unsubscripted type alias to be reduced.
    hint_parent_sane : Optional[HintSane]
        Either:

        * If the passed hint is a **root** (i.e., top-most parent hint of a tree
          of child hints), :data:`None`.
        * Else, the passed hint is a **child** of some parent hint. In this
          case, the **sanified parent type hint metadata** (i.e., immutable and
          thus hashable object encapsulating *all* metadata previously returned
          by :mod:`beartype._check.convert.convmain` sanifiers after
          sanitizing the possibly PEP-noncompliant parent hint of this child
          hint into a fully PEP-compliant parent hint).
    exception_prefix : str
        Human-readable substring prefixing raised exception messages.

    All remaining passed keyword parameters are silently ignored.

    Returns
    -------
    HintSane
        **Sanified type hint metadata** (i.e., :class:`.HintSane` object)
        describing this reduction.

    Raises
    ------
    BeartypeDecorHintPep695Exception
        If this alias contains one or more unquoted relative forward references
        to undefined attributes. Note that this *only* occurs when callers avoid
        beartype import hooks in favour of manually decorating callables and
        classes with the :func:`beartype.beartype` decorator.
    '''
    assert isinstance(hint, HintPep695TypeAlias), (
        f'{repr(hint)} not PEP 695-compliant unsubscripted type alias.')

    # ....................{ RECURSE                        }....................
    #FIXME: *NON-IDEAL.* Ideally, @beartype would actually generate code
    #recursively type-checking recursive hints. However, doing so is *EXTREMELY*
    #non-trivial. Why?
    #
    #Non-triviality is one obvious concern. For each recursive hint, @beartype
    #must now:
    #* Dynamically generate one low-level recursive type-checking function
    #  unique to that recursive hint.
    #* Call each such function in higher-level wrapper functions to type-check
    #  each pith against the corresponding recursive hint.
    #
    #Safety is another obvious concern. Generated code *MUST* explicitly guard
    #against infinitely recursive containers:
    #    >>> infinite_list = []
    #    >>> infinite_list.append(infinite_list)  # <-- gg fam
    #
    #But guarding against infinitely recursive containers requires maintaining a
    #(...waitforit) frozen set of the IDs of all previously type-checked
    #objects, which must then be passed to each dynamically generated recursive
    #type-checking function that type-checks a specific recursive hint.
    #Maintaining these frozen sets then incurs a probably significant space and
    #time complexity hit.
    #
    #In short, it's pretty brutal stuff. For now, simply ignoring recursion
    #strikes us the sanest and certainly simplest approach. *sigh*

    # If this hint is recursive, ignore this hint to avoid infinite recursion.
    #
    # Certainly, various approaches to generating code type-checking recursive
    # hints exists. @beartype currently embraces the easiest, fastest, and
    # laziest approach: just ignore all recursion! Ignorance works wonders.
    if is_hint_recursive(
        hint=hint,  # pyright: ignore
        hint_parent_sane=hint_parent_sane,
        hint_recursable_depth_max=_HINT_PEP695_RECURSABLE_DEPTH_MAX,
    ):
        # print(f'Ignoring recursive PEP 695 unsubscripted type alias {hint} with parent {hint_parent_sane}...')
        return HINT_SANE_RECURSIVE
    # Else, this hint is *NOT* recursive.

    # ....................{ REDUCE                         }....................
    # Underlying type hint to be returned.
    hint_aliased: Hint = None  # pyright: ignore

    # Attempt to...
    try:
        # Reduce this alias to the type hint it lazily refers to. If this alias
        # contains *NO* forward references to undeclared attributes, this
        # reduction *SHOULD* succeed. Let's pretend we mean that.
        #
        # Note that this getter is memoized and thus intentionally called with
        # positional arguments.
        hint_aliased = get_hint_pep695_unsubbed_alias(
            hint, exception_prefix)
    # If doing so raises a builtin "NameError" exception, this alias contains
    # one or more forward references to undeclared attributes. In this case...
    except NameError as exception:
        # Unqualified basename of this alias (i.e., name of the global or local
        # variable assigned to by the left-hand side of this alias).
        hint_name = repr(hint)

        # Fully-qualified name of the third-party module defining this alias.
        hint_module_name = hint.__module__
        # print(f'hint_module_name: {hint_module_name}')

        # Unqualified basename of the next remaining undeclared attribute
        # contained in this alias relative to that module.
        hint_ref_name = get_name_error_attr_name(exception)
        # print(f'hint: {hint}; hint_ref_name: {hint_ref_name}')

        # Raise a human-readable exception describing this issue.
        raise BeartypeDecorHintPep695Exception(
            f'{exception_prefix}PEP 695 type alias "{hint_name}" '
            f'unquoted relative forward reference {repr(hint_ref_name)} in '
            f'module "{hint_module_name}" unsupported outside '
            f'"beartype.claw" import hooks. Consider either:\n'
            f'* Quoting this forward reference in this type alias: e.g.,\n'
            f'      # Instead of an unquoted forward reference...\n'
            f'      type {hint_name} = ... {hint_ref_name} ...\n'
            f'\n'
            f'      # Prefer a quoted forward reference.\n'
            f'      type {hint_name} = ... "{hint_ref_name}" ...\n'
            f'* Applying "beartype.claw" import hooks to '
            f'module "{hint_module_name}": e.g.,\n'
            f'      # In your "this_package.__init__" submodule:\n'
            f'      from beartype.claw import beartype_this_package\n'
            f'      beartype_this_package()'
        ) from exception
    # Else, doing so raised *NO* exceptions, implying this alias contains *NO*
    # forward references to undeclared attributes.

    # ....................{ RETURN                         }....................
    # Sanified metadata to be returned, guarded against infinite recursion.
    hint_sane = make_hint_sane_recursable(
        #FIXME: Document this. Kinda intense, yo. Copy-paste similar comments
        #inside _reduce_hint_overrides() to here, please. *sigh*
        hint_recursable=hint,  # pyright: ignore
        hint_nonrecursable=hint_aliased,
        hint_parent_sane=hint_parent_sane,
    )

    # Return this metadata.
    return hint_sane

# ....................{ PRIVATE ~ constants                }....................
_HINT_PEP695_RECURSABLE_DEPTH_MAX = 1
'''
Value of the optional ``hint_recursable_depth_max`` parameter passed to the
:func:`.is_hint_recursive` tester by the :pep:`695`-compliant reducers defined
above.

This depth ensures that :pep:`695`-compliant type aliases are considered to be
recursive *only* after having been recursed into at most this many times before
(i.e., *only* after having been visited exactly twice, once as a parent alias
and again as a transitive child hint of this parent alias). :pep:`695`-compliant
type aliases typically describe non-trivial recursive data structures conveying
internal semantics that merit deeper recursion. This depth guarantees that.

Consider the following :pep:`695`-compliant subscripted type alias:

.. code-block:: python

   type RecursiveList[T] = list[RecursiveList[T] | T]

Lists satisfying the concrete type alias ``RecursiveList[int]`` contain an
arbitrary number of integers and other lists containing an arbitrary number of
integers, exhibiting this internal structure:

.. code-block:: python

   [42, [79]]
   [42, [79], 83]
   [42, [79], 83, [56, 12]]

Halting recursion at the first expansion of the concrete type alias
``RecursiveList[int]`` to ``list[RecursiveList[int] | int]`` would then reduce
the latter to ``list[HINT_SANE_RECURSIVE | int]``, which reduces to
``list[HINT_SANE_RECURSIVE]``, which reduces to :class:`list`, which clearly
fails to convey the internal semantics of this data structure.

Halting recursion at the second expansion of the concrete type alias
``RecursiveList[int]`` to ``list[RecursiveList[int] | int]`` instead reduces the
latter to ``list[list[RecursiveList[int] | int] | int]``, which reduces to
``list[list[HINT_SANE_RECURSIVE | int] | int]``, reducing to
``list[list[HINT_SANE_RECURSIVE] | int]``, which reduces to ``list[list |
int]`` -- conveying exactly one layer of the internal semantics of this
recursive data structure.
'''



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
**Beartype import hook API.**

This subpackage publishes :pep:`302`- and :pep:`451`-compliant import hooks
enabling external callers to automatically decorate well-typed third-party
packages and modules with runtime type-checking dynamically generated by the
:func:`beartype.beartype` decorator in a single line of code.
'''

# ....................{ TODO                               }....................
#FIXME: [QA] Technically, we're not quite done here. The "beartype.claw" API
#currently silently ignores attempts to subject the "beartype" package itself to
#@beartyping. Ideally, that API should instead raise human-readable exceptions
#when users explicitly attempt to do so when calling either the
#beartype_package() or beartype_packages() functions. After implementing that
#functionality, assert that in our test suite, please.

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To avoid polluting the public module namespace, external attributes
# should be locally imported at module scope *ONLY* under alternate private
# names (e.g., "from argparse import ArgumentParser as _ArgumentParser" rather
# than merely "from argparse import ArgumentParser").
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
from beartype.claw._clawmain import (
    beartype_all as beartype_all,
    beartype_package as beartype_package,
    beartype_packages as beartype_packages,
    beartype_this_package as beartype_this_package,
)
from beartype.claw._package.clawpkgcontext import (
    beartyping as beartyping,
)



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide :pep:`484`-, :pep:`612`-, or :pep:`646`-compliant **type parameter
reducers** (i.e., low-level callables converting arbitrary hints parametrized by
zero or more :pep:`484`-compliant type variables, :pep:`612`-compliant parameter
specifications, and/or :pep:`646`-compliant type variable tuples to lower-level
type hints more readily consumable by :mod:`beartype`).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ TODO                               }....................
#FIXME: [PEP 612] *WOOPS.* Generics can be subscripted by PEP 612-compliant
#parameter specifications: e.g.,
#    # This example literally appears in PEP 696.
#    Ts = TypeVarTuple("Ts")
#    P = ParamSpec("P", default=[float, bool])
#    class Foo(Generic[Ts, P]): ...  # Valid
#
#Brutal. Welp. Let's implement that if somebody actually complains about that,
#please. Doing so will warrant renaming this reducer to "redpep484612646" and
#generalizing everything below. Pretty annoying, but... what can you do? *sigh*

#FIXME: [PEP 696] Handle "PEP 696 – Type Defaults for Type Parameters" under
#Python >= 3.13:
#    https://peps.python.org/pep-0696
#
#This PEP induces edge cases in _make_hint_pep484612646_typearg_to_hint().
#Notably, when the caller passes more type parameters than child hints to that
#factory, we currently silently ignore and thus preserve those "excess" type
#parameters. Under Python >= 3.13, however, we *MUST* instead now:
#* Manually iterate over those any excess *LEADING* type parameters in
#  left-to-right parametrization order. For each such parameter:
#  * If that type parameter defines a default, map that parameter to that
#    default.
#  * Else, halt this iteration immediately.
#
#Note that:
#* Some (or even all) of the above logic may actually already be implicitly
#  supported due to the reduce_hint_pep484612646_typearg() reducer, which now
#  supports PEP 696-compliant defaults.
#* This does apply to both PEP 484-compliant type variables *AND* PEP
#  646-compliant unpacked type variable tuples. Both can be defaulted.
#* This does *NOT* apply to *TRAILING* type parameters. PEP 696 mandates that
#  @beartype should raise an exception if *ANY* trailing type parameter
#  following an unpacked type variable tuple has a default. Just "Ugh!"

# ....................{ IMPORTS                            }....................
from beartype.roar import BeartypeDecorHintPep484612646Exception
from beartype.typing import (
    Optional,
    # TypeVar,
)
from beartype._check.convert._reduce._redrecurse import (
    is_hint_recursive,
    make_hint_sane_recursable,
)
from beartype._check.metadata.hint.hintsane import (
    HINT_SANE_IGNORABLE,
    HINT_SANE_RECURSIVE,
    HintOrSane,
    HintSane,
)
from beartype._check.pep.checkpep484typevar import (
    die_if_hint_pep484_typevar_bound_unbearable)
from beartype._data.error.dataerrmagic import EXCEPTION_PLACEHOLDER
from beartype._data.hint.sign.datahintsigncls import HintSign
from beartype._data.hint.sign.datahintsigns import (
    HintSignTypeVar,
    HintSignPep646TypeVarTupleUnpacked,
)
from beartype._data.kind.datakindiota import SENTINEL
from beartype._data.typing.datatyping import (
    Pep484612646TypeArgUnpacked,
    TuplePep484612646TypeArgsUnpacked,
)
from beartype._data.typing.datatypingport import (
    Hint,
    Pep484612646TypeArgUnpackedToHint,
    TupleHints,
)
from beartype._util.cache.utilcachecall import callable_cached
from beartype._util.error.utilerrraise import reraise_exception_placeholder
from beartype._util.hint.pep.proposal.pep484.pep484typevar import (
    get_hint_pep484_typevar_bounded_constraints_or_none,
    # is_hint_pep484_typevar,
)
from beartype._util.hint.pep.proposal.pep484612646 import (
    die_unless_hint_pep484612646_typearg_unpacked,
    is_hint_pep484612646_typearg_unpacked,
    pack_hint_pep484612646_typearg_unpacked,
)
from beartype._util.hint.pep.proposal.pep646692 import (
    make_hint_pep646_tuple_unpacked_prefix)
from beartype._util.hint.pep.proposal.pep696 import (
    get_hint_pep484612646_typearg_packed_default_or_sentinel)
from beartype._util.hint.pep.utilpepget import (
    get_hint_pep_args,
    get_hint_pep_origin,
    get_hint_pep_typeargs_unpacked,
)
from beartype._util.hint.pep.utilpepsign import get_hint_pep_sign_or_none
from beartype._util.kind.maplike.utilmapfrozen import FrozenDict

# ....................{ REDUCERS                           }....................
def reduce_hint_pep484612646_typearg(
    hint: Hint,
    hint_parent_sane: Optional[HintSane],
    exception_prefix: str,
    **kwargs
) -> HintSane:
    '''
    Reduce the passed :pep:`484`-, :pep:`612`-, or :pep:`646`-compliant **type
    parameter** (i.e., :pep:`484`-compliant type variable, :pep:`612`-compliant
    parameter specification, or :pep:`646`-compliant type variable tuple) to a
    lower-level type hint currently supported by :mod:`beartype`.

    This reducer is intentionally *not* memoized (e.g., by the
    ``callable_cached`` decorator), as reducers cannot be memoized.

    Parameters
    ----------
    hint : Hint
        Type parameter to be reduced.
    hint_parent_sane : Optional[HintSane]
        Either:

        * If the passed hint is a **root** (i.e., top-most parent hint of a tree
          of child hints), :data:`None`.
        * Else, the passed hint is a **child** of some parent hint. In this
          case, the **sanified parent type hint metadata** (i.e., immutable and
          thus hashable object encapsulating *all* metadata previously returned
          by :mod:`beartype._check.convert.convmain` sanifiers after
          sanitizing the possibly PEP-noncompliant parent hint of this child
          hint into a fully PEP-compliant parent hint).
    exception_prefix : str
        Human-readable substring prefixing raised exception messages.

    All remaining keyword-only parameters are silently ignored.

    Returns
    -------
    HintSane
        Either:

        * If this type parameter is **recursive** (i.e., previously transitively
          mapped to itself by a prior call to this reducer), this type parameter
          *must* be ignored to avoid infinite recursion. In this case, the
          :data:`.HINT_SANE_RECURSIVE` singleton.
        * Else if the type parameter lookup table encapsulated by the passed
          sanified parent type hint metadata maps this type parameter to another
          type hint, the latter.
        * Else if this type parameter is both unbounded and unconstrained, this
          type parameter is ignorable. In this case, the
          :data:`.HINT_SANE_IGNORABLE` singleton.
        * Else, this type parameter's lover-level bounds or constraints.
    '''
    # print(f'Reducing PEP 484 type parameter {hint} with parent hint {hint_parent_sane}...')

    # ....................{ PHASE                          }....................
    # This reducer is divided into a series of sequential phases:
    # * Decide whether this type parameter is recursive.
    # * Map this type parameter to its associated target hint by the lookup
    #    table previously associated with this type parameter (if any).
    # * Reduce this type parameter to its default (if any).
    # * Reduce this type parameter to its bounded constraints (if any).
    # * Decide the recursion guard for this type parameter.
    #
    # All phases are non-trivial. The output of each phase is sanified hint
    # metadata (i.e., a "HintSane" object) containing the result of the decision
    # problem decided by that phase.

    # ....................{ PHASE ~ 0 : recurse            }....................
    # If this type parameter is recursive (i.e., previously transitively mapped
    # to itself by a prior call to this reducer), ignore this type parameter to
    # avoid infinite recursion.
    if is_hint_recursive(
        hint=hint,
        hint_parent_sane=hint_parent_sane,
        hint_recursable_depth_max=_TYPEARG_RECURSABLE_DEPTH_MAX,
    ):
        # print(f'Ignoring recursive type parameter {hint} with parent {hint_parent_sane}!')
        return HINT_SANE_RECURSIVE
    # Else, this type parameter is *NOT* recursive.

    # ....................{ PHASE ~ 1 : lookup             }....................
    # Reduced hint to be returned, defaulting to this type parameter.
    hint_reduced = hint

    # If...
    if (
        # This type parameter is not a root hint and thus has a parent hint
        # *AND*...
        hint_parent_sane is not None and
        # A parent hint of this type parameter maps one or more type
        # parameters...
        hint_parent_sane.typearg_to_hint
    ):
        # Type parameter lookup table of this parent hint, localized for
        # usability and negligible efficiency.
        typearg_to_hint = hint_parent_sane.typearg_to_hint

        # If a parent hint of this type parameter maps exactly one type
        # parameter, prefer a dramatically faster and simpler approach.
        if len(typearg_to_hint) == 1:
            # Hint mapped to by this type parameter if one or more parent hints
            # previously mapped this type parameter to a hint *OR* this hint as
            # is otherwise (i.e., if this type parameter is unmapped).
            #
            # Note that this one-liner looks ridiculous, but actually works.
            # More importantly, this is the fastest way to accomplish this.
            hint_reduced = typearg_to_hint.get(hint, hint)  # type: ignore[call-overload]
        # Else, a parent hint of this type parameter mapped two or more type
        # parameters. In this case, fallback to a slower and more complicated
        # approach that avoids worst-case edge cases. This includes recursion in
        # type parameter mappings, which arises in non-trivial class hierarchies
        # involving two or more generics subscripted by two or more type
        # parameters that circularly cycle between one another: e.g.,
        #     from typing import Generic
        #     class GenericRoot[T](Generic[T]): pass
        #
        #     # This directly maps "{T: S}".
        #     class GenericLeaf[S](GenericRoot[S]): pass
        #
        #     # This directly maps "{S: T}", which then combines with the above
        #     # mapping to indirectly map "{S: T, T: S}". Clearly, this indirect
        #     # mapping provokes infinite recursion unless explicitly handled.
        #     GenericLeaf[T]
        else:
            # Type hints previously reduced from this type parameter,
            # initialized to this type parameter.
            hint_reduced_prev = hint

            # Shallow copy of this type parameter lookup table, coerced from an
            # immutable frozen dictionary into a mutable standard dictionary.
            # This enables type parameters reduced by the iteration below to be
            # popped off this copy as a simple (but effective) recursion guard.
            typearg_to_hint_stack = typearg_to_hint.copy()

            #FIXME: [SPEED] *INEFFICIENT.* This has to be done either way, but
            #rather than reperform this O(n) algorithm on every single instance
            #of this type parameter, this algorithm should simply be performed
            #exactly *ONCE* in the
            #reduce_hint_pep484612646_subbed_typeargs_to_hints() reducer. Please
            #refactor this iteration over there *AFTER* the dust settles here.
            #FIXME: Actually, it's unclear how exactly this could be refactored
            #into the reduce_hint_pep484612646_subbed_typeargs_to_hints()
            #reducer. This reduction here only searches for a single typevar in
            #O(n) time. Refactoring this over to
            #reduce_hint_pep484612646_subbed_typeargs_to_hints() would require
            #generalizing this into an O(n**2) algorithm there, probably. Yow!

            # While...
            while (
                # This stack still contains one or more type parameters that
                # have yet to be reduced by this iteration *AND*...
                typearg_to_hint_stack and
                # This hint is still a type parameter...
                is_hint_pep484612646_typearg_unpacked(hint_reduced)
            ):
                # Hint mapped to by this type parameter if one or more parent
                # hints previously mapped this type parameter to a hint *OR* this
                # hint as is (i.e., if this type parameter is unmapped).
                #
                # Note that this one-liner destructively pops this type parameter
                # off this temporary stack to prevent this type parameter from
                # being reduced more than once by an otherwise recursive
                # mapping. Since this stack is local to this reducer, this
                # behaviour is only locally destructive and thus safe.
                hint_reduced = typearg_to_hint_stack.pop(
                    hint_reduced_prev, hint_reduced_prev)  # pyright: ignore

                # If this type parameter maps to itself, this mapping is both
                # ignorable *AND* terminates this reduction.
                if hint_reduced is hint_reduced_prev:
                    break
                # Else, this type parameter does *NOT* map to itself.

                # Map this type parameter to this hint.
                hint_reduced_prev = hint_reduced
        # print(f'...to hint {hint} via type parameter lookup table!')
    # Else, this type parameter is unmapped.

    # ....................{ PHASE ~ 2 : default            }....................
    # If this hint is still an unpacked type parameter...
    if is_hint_pep484612646_typearg_unpacked(hint_reduced):
        # Packed type parameter underlying this unpacked type parameter.
        hint_packed = pack_hint_pep484612646_typearg_unpacked(
            hint=hint_reduced, exception_prefix=exception_prefix)

        # PEP 696-compliant default parametrizing this type parameter if any
        # *OR* the sentinel placeholder otherwise (i.e., if this type parameter
        # has *NO* default).
        hint_default = get_hint_pep484612646_typearg_packed_default_or_sentinel(
            hint=hint_packed, exception_prefix=exception_prefix)

        # If this type parameter has a default, reduce this type parameter to
        # this default.
        if hint_default is not SENTINEL:
            hint_reduced = hint_default
        # Else, this type parameter has *NO* default. In this case, preserve
        # this type parameter as is (for the moment).
    # Else, this hint is no longer a type parameter.

    # ....................{ PHASE ~ 3 : bounds             }....................
    # Sign uniquely identifying this type parameter.
    hint_reduced_sign = get_hint_pep_sign_or_none(hint_reduced)  # pyright: ignore

    # If this hint is still a PEP 484-compliant type variable (e.g., due to
    # either not being mapped by this lookup table *OR* being mapped by this
    # lookup table to yet another type variable)...
    if hint_reduced_sign is HintSignTypeVar:
        # PEP-compliant hint synthesized from all bounded constraints
        # parametrizing this type parameter if any *OR* "None" otherwise (i.e.,
        # if this type parameter is both unbounded *AND* unconstrained).
        #
        # Note this call is passed positional parameters due to memoization.
        hint_reduced = get_hint_pep484_typevar_bounded_constraints_or_none(
            hint_reduced, exception_prefix)  # pyright: ignore

        # If this type parameter is both unbounded *AND* unconstrained, this
        # type parameter is currently *NOT* type-checkable and is thus
        # ignorable. Reduce this type parameter to the ignorable singleton.
        if hint_reduced is None:
            return HINT_SANE_IGNORABLE
        # Else, this type parameter is either bounded *OR* constrained. In
        # either case, preserve this newly synthesized hint.
        # print(f'Reducing PEP 484 type parameter {repr(hint)} to {repr(hint_bound)}...')
        # print(f'Reducing non-beartype PEP 593 type hint {repr(hint)}...')
    # Else, this hint is *NOT* a PEP 484-compliant type variable. If this hint
    # was a type variable, one or more transitive parent hints previously mapped
    # this type parameter to a non-type variable.
    #
    # In any case, only type variables currently accept bounds and constraints;
    # neither PEP 612-compliant parameter specifications *NOR* PEP 646-compliant
    # type variable tuples accept bounds or constraints. Both are unconstrained
    # and thus are currently *NOT* type-checkable and thus are ignorable.
    #
    # If this hint is still a PEP 646-compliant unpacked type variable tuple
    # (e.g., due to either not being mapped by this lookup table *OR* being
    # mapped by this lookup table to yet another unpacked type variable tuple),
    # reduce this unconstrained type parameter to the ignorable singleton.
    elif hint_reduced_sign is HintSignPep646TypeVarTupleUnpacked:
        #FIXME: *HMMM*. This is *ABSOLUTELY* wrong. Type variable tuples aren't
        #like type variables. They actually signify something: notably, that
        #zero or more child hints should be matched. To do so, type variables
        #tuples should be reduced to the PEP 646-compliant unpacked fixed tuple
        #hint "*tuple[object, ...]". Ergo, this should instead resemble here:
        #    return make_hint_pep646_tuple_unpacked_prefix((object, ...))
        #
        #Of course, repeatedly recreating the same hint for *EVERY* single
        #type variable tuple is awful. Instead:
        #* Append a new public global variable to "pep646692" resembling:
        #      HINT_PEP646_TUPLE_UNPACKED_VARIADIC_ANY = make_hint_pep646_tuple_unpacked_prefix(
        #          (object, ...))
        #* Import and return that instead here: e.g.,
        #      return HINT_PEP646_TUPLE_UNPACKED_VARIADIC_ANY
        return HINT_SANE_IGNORABLE
    # Else, this hint is *NOT* a PEP 646-compliant unpacked type variable tuple.
    # In this case, this hint is *NOT* an unconstrained type parameter and thus
    # *NOT* trivially ignorable. Preserve this reduced hint as is.

    # ....................{ PHASE ~ 4 : guard              }....................
    # Decide the recursion guard protecting this possibly recursive type
    # parameter against infinite recursion.
    #
    # Note that this guard intentionally applies to the original unreduced type
    # parameter rather than the newly reduced hint decided by the prior phase.
    # Thus, we pass "hint_recursable=hint" rather than
    # "hint_recursable=hint_reduced".
    hint_sane = make_hint_sane_recursable(
        # The recursable form of this type parameter is its original unreduced
        # form tested by the is_hint_recursive() recursion guard above.
        hint_recursable=hint,
        # The non-recursable form of this type parameter is its new reduced form.
        hint_nonrecursable=hint_reduced,  # pyright: ignore
        hint_parent_sane=hint_parent_sane,
    )

    # ....................{ RETURN                         }....................
    # Return this metadata.
    return hint_sane


#FIXME: Document how PEP 646-compliant unpacked type variable tuples intersect
#with the "Caveats" in the docstring below, please. *megasigh*
def reduce_hint_pep484612646_subbed_typeargs_to_hints(
    # Mandatory parameters.
    hint: Hint,

    # Optional parameters.
    hint_parent_sane: Optional[HintSane] = None,
    exception_prefix: str = '',
) -> HintOrSane:
    '''
    Reduce the passed **subscripted hint** (i.e., derivative hint produced by
    subscripting an unsubscripted hint originally parametrized by one or more
    **type parameters** (i.e., :pep:`484`-compliant type variables or
    :pep:`646`-compliant type variable tuples) with one or more child hints) to
    that unsubscripted hint and corresponding **type parameter lookup table**
    (i.e., immutable dictionary mapping from those same type parameters to those
    same child hints).

    This reducer is intentionally *not* memoized (e.g., by the
    ``callable_cached`` decorator), as reducers cannot be memoized.

    Caveats
    -------
    This reducer does *not* validate these type parameters to actually be type
    parameters. Instead, this function defers that validation to the caller.
    Why? Efficiency, mostly. Avoiding the need to explicitly validate these type
    parameters reduces the underlying mapping operation to a fast one-liner.

    Let:

    * ``hints_typearg`` be the tuple of the zero or more type parameters
      parametrizing the unsubscripted hint underlying the passed subscripted
      hint.
    * ``hints_child`` be the tuple of the zero or more child hints subscripting
      the passed subscripted hint.

    Then this reducer validates the sizes of these tuple to be constrained as:

    .. code-block:: python

       len(hints_typearg) >= len(hints_child) > 0

    Equally, the passed hint *must* be subscripted by at least one child hint.
    For each such child hint, the unsubscripted hint originating this
    subscripted hint *must* be parametrized by a corresponding type parameter.
    The converse is *not* the case, as:

    * For the first type parameter, there also *must* exist a corresponding
      child hint to map to that type parameter.
    * For *all* type parameters following the first, there need *not* exist a
      corresponding child hint to map to that type parameter. Type parameters
      with *no* corresponding child hints are simply silently ignored (i.e.,
      preserved as type parameters rather than mapped to other hints).

    Equivalently:

    * Both of these tuples *must* be **non-empty** (i.e., contain one or more
      items).
    * This tuple of type parameters *must* contain at least as many items as
      this tuple of child hints. Therefore:

      * This tuple of type parameters *may* contain exactly as many items as
        this tuple of child hints.
      * This tuple of type parameters *may* contain strictly more items than
        this tuple of child hints.
      * This tuple of type parameters must *not* contain fewer items than this
        tuple of child hints.

    Parameters
    ----------
    hint : Hint
        Subscripted hint to be inspected.
    hint_parent_sane : Optional[HintSane]
        Either:

        * If the passed hint is a **root** (i.e., top-most parent hint of a tree
          of child hints), :data:`None`.
        * Else, the passed hint is a **child** of some parent hint. In this
          case, the **sanified parent type hint metadata** (i.e., immutable and
          thus hashable object encapsulating *all* metadata previously returned
          by :mod:`beartype._check.convert.convmain` sanifiers after
          sanitizing the possibly PEP-noncompliant parent hint of this child
          hint into a fully PEP-compliant parent hint).

        Defaults to :data:`None`.
    exception_prefix : str, optional
        Human-readable substring prefixing raised exception messages. Defaults
        to the empty string.

    Returns
    -------
    HintOrSane
        Either:

        * If the unsubscripted hint (e.g., :class:`typing.Generic`) originating
          this subscripted hint (e.g., ``typing.Generic[S, T]``) is
          unparametrized by type parameters, that unsubscripted hint as is.
        * Else, that unsubscripted hint is parametrized by one or more type
          parameters. In this case, the **sanified type hint metadata** (i.e.,
          :class:`.HintSane` object) describing this reduction.

    Raises
    ------
    BeartypeDecorHintPep484612646Exception
        If this type hint is unsubscripted.
    BeartypeDecorHintPep484TypeVarViolation
        If one of these type hints violates the bounds or constraints of one of
        these type parameters.
    '''

    # ....................{ LOCALS                         }....................
    # Unsubscripted type alias originating this subscripted hint.
    hint_unsubbed = get_hint_pep_origin(
        hint=hint,
        exception_cls=BeartypeDecorHintPep484612646Exception,
        exception_prefix=exception_prefix,
    )

    #FIXME: [SPEED] Inefficient. This getter internally creates and then
    #discards a full-blown list object just to create this unpacked tuple.
    #Instead, we should:
    #* Call get_hint_pep_typeargs_packed() instead here.
    #* In the _make_hint_pep484612646_typearg_to_hint() factory:
    #  * Detect packed rather than unpacked type variable tuples everywhere.
    #  * Manually pack the detected type variable tuple when mapping this type
    #    variable tuple to another hint: e.g.,
    #        hint_pep646_typevartuple_unpacked = (
    #            make_hint_pep646_typevartuple_unpacked_subbed(
    #                 hint_pep646_typevartuple))
    #        typearg_to_hint[hint_pep646_typevartuple_unpacked] = (
    #            hints_child_excess_tuple_unpacked)

    # Tuple of all unpacked type parameters parametrizing this unsubscripted
    # hint.
    #
    # Note that:
    # * PEP 484-compliant subscripted parametrized generics incorrectly report
    #   being unparametrized, due to outstanding issues in the "typing" module.
    #   Since these issues *ONLY* apply to subscripted rather than unsubscripted
    #   parametrized generics, we strongly prefer the latter for the purposes of
    #   introspecting type parameters:
    #       >>> from beartype.typing import Generic, TypeVar
    #       >>> T = TypeVar('T')
    #       >>> class Ugh(Generic[T]): pass
    #       >>> get_hint_pep_typeargs_packed(Ugh)
    #       (~T,)  # <-- this is good
    #       >>> get_hint_pep_typeargs_packed(Ugh[int])
    #       ()  # <----- THIS IS BAD. wtf, "typing"?
    # * PEP 695-compliant "type" alias syntax superficially appears to
    #   erroneously permit type aliases to be parametrized by non-type
    #   parameters. In truth, "type" syntax simply permits type aliases to be
    #   parametrized by type parameters that ambiguously share the same names as
    #   builtin types -- which then silently shadow those types for the duration
    #   of those aliases:
    #     >>> type muh_alias[int] = float | complex  # <-- *gulp* >>>
    #     muh_alias.__parameters__ (int,)  # <-- doesn't look good so far >>>
    #     muh_alias.__parameters__[0] is int False  # <-- something good finally happened
    hints_typearg = get_hint_pep_typeargs_unpacked(hint_unsubbed)
    # print(f'hints_typearg: {hints_typearg}')

    # Tuple of all child hints subscripting this subscripted hint.
    hints_child = get_hint_pep_args(hint)
    # print(f'hints_child: {hints_child}')

    # ....................{ REDUCE                         }....................
    # Decide the type parameter lookup table for this hint. Specifically, reduce
    # this subscripted hint to:
    # * The semantically useful unsubscripted hint originating this semantically
    #   useless subscripted hint.
    # * The type parameter lookup table mapping all type parameters parametrizing
    #   this unsubscripted hint to all non-type parameter hints subscripting
    #   this subscripted hint.

    # ....................{ REDUCE ~ noop                  }....................
    # If either...
    if (
        # This unsubscripted hint is parametrized by no type parameters *OR*...
        #
        # In this case, *NO* type parameter lookup table can be produced by this
        # reduction. Note this is an uncommon edge case. Examples include:
        # * Parametrizations of the PEP 484-compliant "typing.Generic"
        #   superclass (e.g., "typing.Generic[S, T]"). In this case, the
        #   original unsubscripted "typing.Generic" superclass remains
        #   unparametrized despite that superclass later being parametrized.
        not hints_typearg or
        # This unsubscripted hint is parametrized by the exact same type
        # parameters as this subscripted hint is subscripted by, in which case
        # the resulting type parameter lookup table would uselessly be the
        # identity mapping from each of these type parameters to itself. While
        # an identity type parameter lookup table could trivially be produced,
        # doing so would convey *NO* meaningful semantics and thus be pointless.
        hints_child == hints_typearg
    # Then reduce this subscripted hint to simply this unsubscripted hint, as
    # type parameter lookup tables are then irrelevant.
    ):
        return hint_unsubbed
    # Else, this unsubscripted hint is parametrized by one or more type
    # parameters. In this case, produce a type parameter lookup table mapping
    # these type parameters to child hints subscripting this subscripted hint.

    # ....................{ REDUCE ~ map                   }....................
    # Attempt to...
    try:
        # Type parameter lookup table mapping from each of these type parameters
        # to each of these corresponding child hints.
        #
        # Note that we pass parameters positionally due to memoization.
        typearg_to_hint = _make_hint_pep484612646_typearg_to_hint(
            hint, hints_typearg, hints_child)
    # print(f'Mapped hint {hint} to type parameter lookup table {typearg_to_hint}!')
    # If doing so raises *ANY* exception, reraise this exception with each
    # placeholder substring (i.e., "EXCEPTION_PLACEHOLDER" instance) replaced by
    # an explanatory prefix.
    except Exception as exception:
        reraise_exception_placeholder(
            exception=exception, target_str=exception_prefix)

    # ....................{ REDUCE ~ composite             }....................
    # Sanified metadata to be returned.
    hint_sane: HintSane = None  # type: ignore[assignment]

    # If this hint has *NO* parent, this is a root hint. In this case...
    if hint_parent_sane is None:
        # Metadata encapsulating this hint and type parameter lookup table.
        hint_sane = HintSane(
            hint=hint_unsubbed, typearg_to_hint=typearg_to_hint)
    # Else, this hint has a parent. In this case...
    else:
        # If the parent hint is also associated with a type parameter lookup
        # table...
        if hint_parent_sane.typearg_to_hint:
            # Full type parameter lookup table merging the table associated this
            # parent hint with the table just decided above for this child hint,
            # efficiently defined as...
            typearg_to_hint = (
                # The type parameter lookup table describing all transitive
                # parent hints of this hint with...
                hint_parent_sane.typearg_to_hint |  # type: ignore[operator]
                # The type parameter lookup table describing this hint.
                #
                # Note that this table is intentionally the second rather than
                # first operand of this "|" operation, efficiently ensuring that
                # type parameters mapped by this hint take precedence over type
                # parameters mapped by transitive parent hints of this hint.
                typearg_to_hint
            )
        # Else, the parent hint is associated with *NO* such table.

        # Metadata encapsulating this hint and type parameter lookup table,
        # while "cascading" any other metadata associated with this parent hint
        # (e.g., recursable hint IDs) down onto this child hint as well.
        hint_sane = hint_parent_sane.permute_sane(
            hint=hint_unsubbed, typearg_to_hint=typearg_to_hint)

    # ....................{ RETURN                         }....................
    # print(f'Reduced subscripted hint {repr(hint)} to unsubscripted hint metadata {repr(hint_sane)}.')

    # Return this metadata.
    return hint_sane

# ....................{ PRIVATE ~ constants                }....................
_TYPEARG_RECURSABLE_DEPTH_MAX = 0
'''
Value of the optional ``hint_recursable_depth_max`` parameter passed to the
:func:`.is_hint_recursive` tester by the :func:`.reduce_hint_pep484_typevar`
reducer.

This depth ensures that :pep:`484`- or :pep:`646`-compliant type parameters are
considered to be recursive *only* after having been recursed into at most this
many times before (i.e., *only* after having been visited exactly once as a
child hint of an arbitrary parent hint). By definition, type parameters are
guaranteed to *never* be parent hints. Ergo, recursing into type parameters
exactly once suffices to expand exactly one nesting level of non-trivial
recursive data structures. This mirrors the restrained recursion allowed by
other reducers.

Consider :pep:`695`-compliant type aliases, for example. Unlike type parameters,
type aliases are guaranteed to *always* be parent hints. The value of the
optional ``hint_recursable_depth_max`` parameter passed to the
:func:`.is_hint_recursive` tester by :pep:`695`-specific reducers is thus one
greater than the value of this global. Nonetheless, the real-world effect is
exactly the same: exactly one nesting level of type aliases is expanded.

Consider the following :pep:`484`-compliant generic recursively subscripted by
itself via a :pep:`484`-compliant type variable ``T``:

.. code-block:: python

   class GenericList[T](): ...
   RecursiveGenericList = GenericList[GenericList]

Instances of this generic satisfying the type hint ``RecursiveGenericList``
contain an arbitrary number of other instances of this generic, exhibiting this
internal structure:

.. code-block:: python

   recursive_generic_list = GenericList()
   recursive_generic_list.append(recursive_generic_list)

Halting recursion at the first expansion of the type parameter ``T`` then
reduces the type hint ``RecursiveGenericList`` to
``GenericList[GenericList[T]]``, , which reduces to
``GenericList[GenericList[HINT_SANE_RECURSIVE]]``, , which reduces to
``GenericList[GenericList]`` -- conveying exactly one layer of the internal
semantics of this recursive data structure.
'''

# ....................{ PRIVATE ~ raisers                  }....................
def _die_unless_hint_pep484_typevar_bound_bearable(
    hint_child: Hint,
    hint_typearg: Pep484612646TypeArgUnpacked,
    hint_typearg_sign: Optional[HintSign],
) -> None:
    '''
    Raise an exception if either:

    * The passed type parameter is a :pep:`484`-compliant type variable *and*
      the passed child hint violates this type variable's bounds and/or
      constraints.
    * The passed type parameter is *not* a :pep:`484`-compliant type variable.

    Parameters
    ----------
    hint_child : Hint
        Child hint to be inspected.
    hint_typearg : Pep484612646TypeArgUnpacked
        Type parameter to be inspected.
    hint_typearg_sign : Optional[HintSign]
        Sign uniquely identifying this type parameter.

    Raises
    ------
    BeartypeDecorHintPep484612646Exception
        If this type parameter is *not* a :pep:`484`-compliant type variable.
    BeartypeDecorHintPep484TypeVarViolation
        If this type parameter is a :pep:`484`-compliant type variable *and*
        this child hint violates this type variable's bounds and/or constraints.
    '''

    # If this type parameter is a PEP 484-compliant type variable...
    if hint_typearg_sign is HintSignTypeVar:
        # If this child hint violates this type variable's bounds and/or
        # constraints, raise an exception.
        die_if_hint_pep484_typevar_bound_unbearable(
            hint=hint_child,
            typevar=hint_typearg,  # type: ignore[arg-type]
            exception_prefix=EXCEPTION_PLACEHOLDER,
        )
        # Else, this child hint satisfies this type variable's bounds and/or
        # constraints.
    # Else, this type parameter is *NOT* a PEP 484-compliant type variable.
    # Ergo, this type parameter is *NOT* an unpacked type parameter! *ROAR*.
    #
    # Note that this should *NEVER* occur. Python itself syntactically
    # guarantees *ALL* child hints parametrizing a PEP-compliant subscripted
    # hint to be unpacked type parameters. Nonetheless, the caller is under
    # no such constraints. To guard against dev bitrot, we validate this.
    else:
        die_unless_hint_pep484612646_typearg_unpacked(
            hint=hint_typearg, exception_prefix=EXCEPTION_PLACEHOLDER)  # pyright: ignore

# ....................{ PRIVATE ~ factories                }....................
#FIXME: Unit test that this reducer reduces PEP 646-compliant unpacked type
#variable tuples, please. *sigh*
@callable_cached
def _make_hint_pep484612646_typearg_to_hint(
    hint: Hint,
    hints_typearg: TuplePep484612646TypeArgsUnpacked,
    hints_child: TupleHints,
) -> Pep484612646TypeArgUnpackedToHint:
    '''
    Type parameter lookup table mapping from the passed :pep:`484`- or
    :pep:`646`-compliant **type parameters** (i.e., :pep:`484`-compliant type
    variables or :pep:`646`-compliant type variable tuples) to the associated
    passed type hints as key-value pairs of this table.

    This getter is memoized for efficiency. Notably, this getter creates and
    returns a dictionary mapping each type parameter in the passed tuple of type
    parameters to the associated type hint in the passed tuple of type hints
    with the same 0-based tuple index as that type parameter.

    Parameters
    ----------
    hint: Hint
        Parent hint presumably both subscripted by these child hints. This
        parent hint is currently only used to generate human-readable exception
        messages in the event of fatal errors.
    hints_typearg : TuplePep484612646TypeArgsUnpacked
        Tuple of one or more child type parameters originally parametrizing the
        origin underlying this parent hint.
    hints_child : TupleHints
        Tuple of zero or more child type hints subscripting this parent hint,
        which those type parameters map to.

    Returns
    -------
    Pep484612646TypeArgUnpackedToHint
        Type parameter lookup table mapping these type parameters to these child
        hints.

    Raises
    ------
    BeartypeDecorHintPep484612646TypeArgException
        If either:

        * This tuple of type parameters is empty.
        * This tuple of type hints is empty.
        * This tuple of type hints contains more items than this tuple of type
          parameters.
    BeartypeDecorHintPep484TypeVarViolation
        If one of these type hints violates the bounds or constraints of one of
        these type parameters.
    '''
    assert isinstance(hints_typearg, tuple), (
        f'{repr(hints_typearg)} not tuple.')
    assert isinstance(hints_child, tuple), (
        f'{repr(hints_child)} not tuple.')
    # print(f'hints_typearg: {hints_typearg}')
    # print(f'hints_child: {hints_child}')

    # ....................{ PREAMBLE                       }....................
    # Number of passed type parameters and child hints respectively.
    hints_typearg_len = len(hints_typearg)
    hints_child_len = len(hints_child)

    # If *NO* type parameters were passed, raise an exception.
    if not hints_typearg_len:
        raise BeartypeDecorHintPep484612646Exception(
            f'{EXCEPTION_PLACEHOLDER}type hint {repr(hint)} '
            f'parametrized by no type parameters (i.e., '
            f'PEP 484 type variables, '
            f'PEP 612 unpacked parameter specifications, or '
            f'PEP 646 unpacked type variable tuples).'
        )
    # Else, one or more type parameters were passed.

    # ....................{ LOCALS                         }....................
    # Type parameter lookup table to be returned.
    typearg_to_hint: Pep484612646TypeArgUnpackedToHint = {}

    # 0-based index of the current type parameter *AND* corresponding child hint
    # of the passed tuples visited by the "while" loop below.
    hints_index_curr = 0

    # 0-based index of the last type parameter *AND* corresponding child hint of
    # the passed tuples to be visited by the "while" loop below, intentionally
    # defined as the maximum number of type parameters or child hints. Why?
    # Because two relevant edge cases then arise:
    # * If more type parameters than child hints were passed, this calculation
    #   induces the "while" loop below to silently ignore those trailing type
    #   parameters that lack corresponding trailing child hints -- exactly as
    #   required and documented by the above docstring. This case is valid! \o/
    # * If more child hints than type parameters were passed, two sub-edge cases
    #   now arise:
    #   * If one of these type parameters is a PEP 646-compliant unpacked type
    #     variable tuple, this parent edge case is valid. Although more child
    #     hints than type parameters were passed, a single unpacked type
    #     variable tuple consumes zero or more excess child hints (i.e., child
    #     hints *NOT* already consumed by one or more trailing PEP 484-compliant
    #     type variables).
    #   * If none of these type parameters is a PEP 646-compliant unpacked type
    #     variable tuple, this parent edge case is invalid. Validation below
    #     will subsequently detect this invalid case and raise an exception.
    hints_index_last = max(hints_typearg_len, hints_child_len) - 1

    # Either:
    # * If the passed tuple of type parameters contains one or more PEP
    #   646-compliant unpacked type variable tuples, the 0-based index of
    #   the first such type parameter.
    # * Else, "None".
    #
    # This local variable is used to detect and track such type parameters.
    hints_pep646_typevartuple_index: Optional[int] = None

    # ....................{ PHASE 1 ~ pep 484 : typevar    }....................
    # In this first phase, we map all *LEADING* PEP 484-compliant type variables
    # parametrizing the beginning of this parent subscripted hint until
    # discovering the first PEP 646-compliant type variable tuple (if any)
    # parametrizing the middle of this parent subscripted hint. Notably:
    # * Type variables only map to and thus "consume" a single child hint.
    #   Mapping leading type variables is trivial.
    # * Type variable tuples map to and thus "consume" zero or more child hints.
    #   Mapping both type variable tuples *AND* the trailing type variables that
    #   follow them is comparatively non-trivial. Special care is warranted.

    # While the 0-based index of the current leading type variable does *NOT*
    # exceed that of the last child hint to be visited by this loop, one or
    # more leading type variables remain unvisited. In this case...
    while hints_index_curr <= hints_index_last:
        # print(f'Visiting leading type parameter index {hints_index_curr} <= {hints_index_last}...')

        # If the 0-based index of this leading type variable exceeds that of the
        # last type variable, more child hints than type parameters were passed.
        # Why? Because, if fewer or the same number of child hints as type
        # parameters were passed, then this index would *NEVER* exceed that of
        # the last type variable. But this index exceeded that of the last type
        # variable! The converse must thus be true. Raise an exception.
        #
        # Note that it is conditionally valid for the caller to pass more child
        # hints than type parameters if one of the previously visited type
        # parameters is a PEP 646-compliant unpacked type variable tuple, which
        # would have then consumed all excess child hints. However, one of these
        # type parameters is *NOT* an unpacked type variable tuple. Why? Because
        # if one of these type parameters was such a tuple, then this "while"
        # loop would have already been prematurely terminated by the "break"
        # statement below. Clearly, though, this "while" loop is still
        # iterating! Ergo, the converse must yet again be true.
        if hints_index_curr >= hints_typearg_len:
            raise BeartypeDecorHintPep484612646Exception(
                f'{EXCEPTION_PLACEHOLDER}type hint {repr(hint)} '
                f'number of subscripted child hints {hints_child_len} exceeds '
                f'number of parametrized type parameters {hints_typearg_len} '
                f'(i.e., {hints_child_len} > {hints_typearg_len}).'
            )
        # Else, the 0-based index of this type variable does *NOT* exceed that
        # of the last child hint.

        # Current leading type parameter.
        hint_typearg = hints_typearg[hints_index_curr]

        # Sign uniquely identifying this type parameter.
        hint_typearg_sign = get_hint_pep_sign_or_none(hint_typearg)  # pyright: ignore
        # print(f'Visiting leading type parameter {hint_typearg} of {hint_typearg_sign}...')

        # If this type parameter is a PEP 646-compliant unpacked type variable
        # tuple...
        if hint_typearg_sign is HintSignPep646TypeVarTupleUnpacked:
            # 0-based index of this unpacked type variable tuple.
            hints_pep646_typevartuple_index = hints_index_curr

            # Immediately halt all further visitation of type parameters and
            # child hints by this iteration. Unpacked type variable tuples
            # greedily consume all remaining child hints and thus warrant
            # special handling below.
            break
        # Else, this type parameter is *NOT* an unpacked type variable tuple.
        #
        # If the 0-based index of this trailing child hint exceeds that of the
        # last child hint, more type parameters than child hints were passed by
        # similar logic as above. In this case...
        #
        # Note that this condition is intentionally tested here *AFTER* ensuring
        # that this parent hint is parametrized by *NO* unpacked type variable
        # tuple. Why? Because it is conditionally valid for the caller to pass
        # exactly *ONE* more type parameter than child hints if one of those
        # type parameters is an unpacked type variable tuple, which would have
        # then consumed *ZERO* child hints and thus effectively *NOT* have been
        # passed at all. Testing this condition *AFTER* detecting an unpacked
        # type variable tuple enables that type parameter to be correctly mapped
        # to the empty tuple below.
        elif hints_index_curr >= hints_child_len:
            # If *NO* child hints were passed, raise an exception.
            #
            # Note that this condition is intentionally tested here *AFTER*
            # ensuring that this parent hint is parametrized by *NO* unpacked
            # type variable tuple. Why? Because it is conditionally valid for
            # the caller to pass exactly *ZERO* child hints and *ONE* unpacked
            # type variable tuple, which would have then consumed *ZERO* child
            # hints and thus effectively *NOT* have been passed at all.
            if not hints_child_len:
                raise BeartypeDecorHintPep484612646Exception(
                    f'{EXCEPTION_PLACEHOLDER}type hint {repr(hint)} '
                    f'subscripted by no child type hints but '
                    f'parametrized by PEP 484 type variables '
                    f'{repr(hints_typearg)} necessarily matching '
                    f'at least one child hint.'
                )
            # Else, one or more child hints were passed.

            # Immediately halt all further visitation of type parameters and
            # child hints by this iteration. All remaining type variables will
            # be silently ignored and thus preserved as is *WITHOUT* being
            # mapped -- a valid edge case.
            break
        # Else, the 0-based index of this child hint does *NOT* exceed that
        # of the last child hint.

        # Current leading child hint.
        hint_child = hints_child[hints_index_curr]
        # print(f'Mapping typearg {typearg} -> hint {hint}...')
        # print(f'is_hint_nonpep_type({hint})? {is_hint_nonpep_type(hint, False)}')

        # Raise an exception if either:
        # * This type parameter is a PEP 484-compliant type variable *AND*
        #   this child hint violates this type variable's bounded constraints.
        # * This type parameter is *NOT* a PEP 484-compliant type variable.
        _die_unless_hint_pep484_typevar_bound_bearable(
            hint_child=hint_child,
            hint_typearg=hint_typearg,
            hint_typearg_sign=hint_typearg_sign,
        )
        # Else, this type parameter is a PEP 484-compliant type variable *AND*
        # this child hint satisfies this type variable's bounded constraints.

        # Map this type variable to this hint with an optimally efficient
        # one-liner, silently overwriting any prior such mapping of this type
        # variable by either this call or a prior call of this function.
        typearg_to_hint[hint_typearg] = hint_child
        # print(f'Mapping {hint_typearg} to {hint_child}...')

        # Iterate the 0-based index of the current type parameter *AND*
        # corresponding child hint to be visited by the next loop iteration.
        hints_index_curr += 1

    # If a PEP 646-compliant unpacked type variable tuple was visited by the
    # "while" loop above...
    if hints_pep646_typevartuple_index is not None:
        # ....................{ PHASE 2 ~ pep 484 : typevar}....................
        # In this next phase, we map all *TRAILING* PEP 484-compliant type
        # variables parametrizing the ending of this parent subscripted hint.
        # These variables follow the first (and ideally only) PEP 646-compliant
        # type variable tuple parametrizing the middle of this parent
        # subscripted hint.
        #
        # The "while" loop above iterates forward over type variables starting
        # at the first type variable, as is customary for most iteration. In
        # contrast, the "while" loop below iterates backward over type variables
        # starting at the last type variable. Why? Because type variable tuples
        # greedily consume all remaining child hints that have yet to be
        # consumed by a trailing type variable. To decide which child hints
        # remain to be consumed by a type variable tuple requires that we first
        # consume as many trailing child hints as possible by as many trailing
        # type variables exist. Whatever child hints remain are then apportioned
        # to the type variable tuple.
        # print(f'hints_pep646_typevartuple_index: {hints_pep646_typevartuple_index}')

        # This unpacked type variable tuple.
        hint_pep646_typevartuple = hints_typearg[
            hints_pep646_typevartuple_index]

        # 0-based index of the current trailing type parameter visited by the
        # "while" loop below, initialized to that of the last type parameter.
        hints_pep646_typearg_index_curr = hints_typearg_len - 1
        # print(f'hints_pep646_typearg_index_curr: {hints_pep646_typearg_index_curr}')

        # 0-based index of the current trailing child hint visited by the
        # "while" loop below, initialized to that of the last child hint.
        hints_pep646_child_index_curr = hints_child_len - 1
        # print(f'hints_pep646_child_index_curr: {hints_pep646_child_index_curr}')

        # 0-based index of the first trailing type parameter to be visited last
        # by the "while" loop below, initialized to that of the first type
        # parameter following this unpacked type variable tuple. This and all
        # following type parameters have yet to consume any child hints and thus
        # *MUST* thus consume those child hints before this unpacked type
        # variable tuple if given a fallback chance to greedily do so below.
        # Every type parameter necessarily consumes exactly one child hint and
        # thus assumes precedence over any unpacked type variable tuple, which
        # permissively consumes zero or more child hints.
        hints_pep646_typearg_index_first = hints_pep646_typevartuple_index + 1
        # print(f'hints_pep646_typearg_index_first: {hints_pep646_typearg_index_first}')

        # 0-based index of the first trailing child hint to be visited last by
        # the "while" loop below, initialized to that of the first child hint
        # following the last *LEADING* child hint consumed by the "while" loop
        # above. This and all following child hints have yet to be consumed and
        # thus remain available for consumption by trailing type parameters.
        hints_pep646_child_index_first = hints_index_curr
        # print(f'hints_pep646_child_index_first: {hints_pep646_child_index_first}')

        # While the 0-based index of the current trailing type parameter still
        # follows that of the unpacked type variable tuple to *NOT* be visited
        # by this loop, one or more trailing type parameters remain unconsumed.
        # In this case...
        while (
            hints_pep646_typearg_index_curr >=
            hints_pep646_typearg_index_first
        ):
            # print(f'Visiting trailing type parameter index {hints_pep646_typearg_index_curr}...')
            # print(f'>= {hints_pep646_typearg_index_first}...')

            # Current trailing type parameter.
            hint_typearg = hints_typearg[hints_pep646_typearg_index_curr]

            # Sign uniquely identifying this type parameter.
            hint_typearg_sign = get_hint_pep_sign_or_none(hint_typearg)  # pyright: ignore
            # print(f'Visiting trailing type parameter {hint_typearg} of {hint_typearg_sign}...')

            # If this type parameter is a second PEP 646-compliant unpacked type
            # variable tuple, raise an exception. PEP 646 mandates that generics
            # be parametrized by at most one unpacked type variable tuple.
            if hint_typearg_sign is HintSignPep646TypeVarTupleUnpacked:
                raise BeartypeDecorHintPep484612646Exception(
                    f'{EXCEPTION_PLACEHOLDER}type hint {repr(hint)} '
                    f'parametrized by PEP 646-noncompliant type parameters '
                    f'{repr(hints_typearg)} containing two or more PEP 646 '
                    f'unpacked type variable tuples, including:\n'
                    f'* {repr(hint_pep646_typevartuple)} at index '
                    f'{hints_pep646_typevartuple_index}.\n'
                    f'* {repr(hint_typearg)} at index '
                    f'{hints_pep646_typearg_index_curr}.\n'
                )
            # Else, this type parameter is *NOT* a second PEP 646-compliant
            # unpacked type variable tuple.
            #
            # If the 0-based index of this trailing child hint precedes that of
            # the first trailing child hint, more type parameters than child
            # hints were passed by similar logic as above. In this case,
            # immediately halt all further visitation of type parameters and
            # child hints by this iteration. All remaining type variables will
            # be silently ignored and thus preserved as is *WITHOUT* being
            # mapped -- a valid edge case.
            #
            # Note that, unlike above, it is valid for *NO* child hints to be
            # passed. Why? Because this parent hint is parametrized by an
            # unpacked type variable tuple, which will then consume *ZERO* child
            # hints below.
            elif (
                hints_pep646_child_index_curr <
                hints_pep646_child_index_first
            ):
                # print('Ignoring all trailing excess type parameters...')
                break
            # Else, the 0-based index of this trailing child hint exceeds or is
            # equal to that of the first trailing child hint.

            # Current trailing child hint.
            hint_child = hints_child[hints_pep646_child_index_curr]

            # Raise an exception if either:
            # * This type parameter is a PEP 484-compliant type variable *AND*
            #   this child hint violates this type variable's bounds.
            # * This type parameter is *NOT* a PEP 484-compliant type variable.
            _die_unless_hint_pep484_typevar_bound_bearable(
                hint_child=hint_child,
                hint_typearg=hint_typearg,
                hint_typearg_sign=hint_typearg_sign,
            )
            # Else, this type parameter is a PEP 484-compliant type variable
            # *AND* this child hint satisfies this type variable's bounds.

            # Map this type variable to this child hint with an efficient
            # one-liner, overwriting any prior such mapping of this type
            # variable by either this call or a prior call of this function.
            typearg_to_hint[hint_typearg] = hint_child
            # print(f'Mapping {hint_typearg} to {hint_child}...')

            # Iterate the 0-based indices of the current type parameter *AND*
            # corresponding child hint to be visited by the next loop iteration.
            hints_pep646_typearg_index_curr -= 1
            hints_pep646_child_index_curr -= 1

        # ....................{ PHASE 3 ~ pep 646 : tuple  }....................
        # In this next phase, we map the sole PEP 646-compliant unpacked type
        # variable tuple parametrizing the middle of this parent subscripted
        # hint. This unpacked type variable tuple:
        # * Follows the leading PEP 484-compliant type variables mapped by the
        #   first phase above.
        # * Precedes the trailing PEP 484-compliant type variables mapped by the
        #   second phase above.
        #
        # If...
        if (
            # The 0-based index of the current type parameter to be consumed is
            # that of the middle unpacked type variable tuple *AND*...
            #
            # In this case, *ALL* trailing type parameters were consumed. The
            # only remaining type parameter to be consumed is the middle
            # unpacked type variable tuple.
            (
                hints_pep646_typearg_index_curr ==
                hints_pep646_typevartuple_index
            ) and
            # The 0-based index of the current trailing child hint to be
            # consumed is or exceeds that of the last *LEADING* child hint
            # consumed by the "while" loop far above...
            #
            # Two cases arise. The 0-based index of the current trailing child
            # hint to be consumed is either:
            # * Exactly that of the last *LEADING* child hint. Then *ALL*
            #   trailing child hints were consumed by the "while" loop above.
            #   Zero trailing child hints remain to be consumed by this unpacked
            #   type variable tuple -- a valid edge case.
            # * Greater than that of the last *LEADING* child hint. Then one or
            #   more trailing child hints were *NOT* consumed by the "while"
            #   loop above. One or more trailing child hints remain to be
            #   consumed by this unpacked type variable tuple -- also a valid
            #   use case.
            #
            # In either case, zero or more trailing child hints were *NOT*
            # consumed by the "while" loop above. Since an unpacked type
            # variable tuple permissively consumes zero or more child hints,
            # this unpacked type variable tuple should now do so.
            (
                hints_pep646_child_index_curr >=
                hints_pep646_child_index_first - 1
            )
        ):
            # print('Mapping unpacked type variable tuple to excess child hints...')

            # Tuple of the zero or more excess child hints *NOT* already
            # consumed above by a trailing type variable, defined here as the
            # slice of the tuple of the zero or more trailing child hints...
            #
            # Note that raw tuples are, by definition, PEP-noncompliant. Ergo,
            # this tuple is PEP-noncompliant as well and thus unsuitable for
            # being directly mapped to this unpacked type variable tuple.
            hints_child_excess = hints_child[
                # Starting at the index of the first trailing child hint
                # *AND*...
                hints_pep646_child_index_first:
                # Ending at the index of the last excess trailing child hint.
                # Note that slice syntax requires incrementing the last index,
                # which it treats as analogous to a length. Specifically, slice
                # syntax ignores the item at the last index.
                #
                # Oh, Python... You sweet summer child.
                hints_pep646_child_index_curr + 1
            ]
            # print(f'hints_child_excess: {hints_child_excess}')

            # Target hint to which this unpacked type variable will be mapped
            # below, synthesized from this PEP-noncompliant tuple slice into a
            # PEP-compliant type hint.
            hint_pep646_typevartuple_target: Hint = SENTINEL  # type: ignore[assignment]

            # If exactly one child hint has *NOT* already consumed, directly map
            # this unpacked type variable to this child hint rather than mapping
            # this unpacked type variable to the 1-tuple containing this child
            # hint.
            #
            # Note that this is *NOT* merely a negligible optimization, although
            # this is technically that. This edge case is required to correctly
            # map unpacked type variable tuples to other unpacked type variable
            # tuples *WITHOUT* triggering infinite recursion. Why? Because the
            # reduce_hint_pep484612646_typearg() function reduces unpacked type
            # variable tuples that have been previously mapped to unpacked type
            # variable tuples with iteration over the returned "typearg_to_hint"
            # dictionary. For efficiency and simplicity, that iteration only
            # supports direct mappings. That iteration does *NOT* support an
            # unpacked type variable tuple mapping to a 1-tuple containing an
            # unpacked type variable tuple. Consider an example that would
            # induce infinite recursion (unless explicitly handled):
            #     from typing import Generic, TypeVarTuple
            #
            #     Ts = TypeVarTuple('Ts')
            #     Tt = TypeVarTuple('Tt')
            #
            #     class RootGeneric(    Generic[*Ts]): pass
            #     class StemGeneric(RootGeneric[*Tt]): pass
            #     class LeafGeneric(StemGeneric[*Ts]): pass#
            if len(hints_child_excess) == 1:
                hint_pep646_typevartuple_target = hints_child_excess[0]
            # Else, either zero or two or more child hints have *NOT* already
            # consumed. In either case, only a tuple of child hints correctly
            # expresses this excessive mapping.
            else:
                #FIXME: [SPEED] Inefficient. This factory internally creates and
                #then discards a full-blown list object just to create this
                #unpacked tuple. Instead, we should:
                #* Define a new make_hint_pep646_tuple_unpacked_pure() factory,
                #  dynamically creating and returning a new semantically equivalent
                #  pure-Python "typing.Unpack[tuple[...]]" hint.
                #* Call that factory here instead.

                # PEP 646-compliant unpacked tuple hint of the form
                # "*tuple[hints_child_excess[0], ..., hints_child_excess[N]]".
                hint_pep646_typevartuple_target = (
                    make_hint_pep646_tuple_unpacked_prefix(hints_child_excess))
            # print(f'hint_pep646_typevartuple_target: {hint_pep646_typevartuple_target}')

            # Map this unpacked type variable tuple to this hint with a
            # one-liner, overwriting any prior such mapping of this type
            # variable by either this call or a prior call of this function.
            #
            # Note that PEP standards support mapping type variable tuples
            # *ONLY* to other type variable tuples or unpacked tuple hints.
            # Since the former fails to apply here, only the latter applies.
            # Consider PEP 696, for example, which explicitly states:
            #     TypeVarTuple defaults are defined using the same syntax as
            #     TypeVars but use an unpacked tuple of types instead of a
            #     single type or another in-scope TypeVarTuple (see Scoping
            #     Rules).
            #
            #     DefaultTs = TypeVarTuple(
            #         "DefaultTs", default=Unpack[tuple[str, int]])
            #
            # In the terminology of this factory function, the above example
            # would effectively add this mapping to "typearg_to_hint":
            #     {DefaultTs: *tuple[str, int]], ...}
            typearg_to_hint[hint_pep646_typevartuple] = (
                hint_pep646_typevartuple_target)
    # Else, a PEP 646-compliant unpacked type variable tuple was *NOT* visited
    # by the "while" loop above. In this case, that loop has already
    # successfully visited all PEP 484-compliant type variables and thus all
    # type parameters parametrizing this parent subscripted hint. We are done!

    # ....................{ RETURN                         }....................
    # Return this table, coerced into an immutable frozen dictionary.
    return FrozenDict(typearg_to_hint)
